#ifndef BUFFERMANAGER_H_
#define BUFFERMANAGER_H_

#pragma once

using namespace std;

#include "../DataHeader/DataHeader.h"
#include "../DataPage/DataPage.h"

#include "../Buffer/BufferPool.h"
#include "../Buffer/BufferInfoTable.h"
#include "../Buffer/BufferInfoTableState.h"
#include "../Buffer/BITSnapShot.h"
#include "../Buffer/BlockForInsert.h"

#include "../Managers/CatalogManager.h"
#include "../Managers/StorageManager.h"
#include "../Managers/CatalogManager.h"
#include "../Managers/MessageManager.h"

// helper classes ..
#include "../Aggregations/FreeAndFlushBlockIndexAggregation.h"
#include "../Aggregations/HeaderEntryResultsAggregation.h"
#include "../Aggregations/LoadAndInsertIntoPartialAggregation.h"
#include "../Aggregations/InsertIntoEmptyAggregation.h"
#include "../Aggregations/InsertIntoPartialAggregation.h"


#include <string>
#include <vector>
#include <list>
#include <map>
#include <iostream>
#include <fstream>


class BufferManager {
	
public:
	
	BufferManager( int bufferPoolSize, StorageManager* sm_ );
	~BufferManager(void);
	
	
	// FUTURE: make "48" byte block size parameterized 
	//         ( page-size user-defined ) ..
	
		
	// member methods ..
	
	StorageManager* getStorageManager();
	
	CatalogManager* getCatalogManager();
	
	MessageManager* getMessageManager();
	
	
	
	// utility method (workaround) ..
	void purgeLocksMap( map< string, vector<int>* >* locks );
	
	
	// flush / load methods ..
	bool writeBlock( DataPage* dataPage, map< string, vector<int>* >* locks );
	bool loadBlock( string tableName, DataPage* dataPage, map< string, vector<int>* >* locks, bool lockPage );
	
	// remove / add lock methods ..
	void removeLock( string tableName, int pageID );
	void addLock( string tableName, int pageID );
	
	// searches "BP" for a block available for flushing and insert (LRU, pin_count=0) ..
	int findLRUAvailableBlockForFlushing();
	
	// "dirty" pages are "latest" copies, or pages that have not yet been 
	// written to file, and these are the pageIDs to return ..
	vector<int>* getDirtyBufferPoolBlocks();
	
	// searches "BP" for a block with a free slot for insert ..
	//int findBlockForInsert( string tableNameX, int numberOfAttributes );
	BlockForInsert* findBlockForInsert( string tableNameX, int numberOfAttributes );
	
	// searches "BP" for an empty block to load a page-from-file into ..
	int findBlockForLoad();
	
	// helper method used by "insertRecordIntoBlock()" ..
	FreeAndFlushBlockIndexAggregation* getFreeAndFlushBlockIndexProcessing( int newFreeBlockIndex );
	
	// helper method used by "insertRecordIntoBlock()" ..
	bool flushBPBlock( int flushBlockIndex, DataHeader* hp );
	
	// helper method used by "insertRecordIntoBlock()" ..
	HeaderEntryResultsAggregation* getHeaderEntryResults( DataHeader* header, vector<int>* bpPageIDs );
	
	// helper method by "insertRecordIntoBlock()" ..
	void loadAndInsertIntoPartial( LoadAndInsertIntoPartialAggregation* laiipp );
	
	// helper method by "insertRecordIntoBlock()" ..
	void insertIntoEmpty( InsertIntoEmptyAggregation* iiep );
	
	// helper method by "insertRecordIntoBlock()" ..
	void insertIntoPartial( InsertIntoPartialAggregation* iipp );
	
	// inserts a record into a block ..
	bool insertRecordIntoBlock( string tableName, vector<int>* insertValues, int numberOfAttributes );
	
	// write contents of buffer pool to file ..
	//bool commitBufferPool();
	bool commitBufferPool( vector<string>* tableNames );
	
	// returns a DataPage in the Buffer Pool ..
	DataPage* returnBPBlock( int bitIndex );
	
	// returns a DataPage from the file ..
	DataPage* returnFilePage( string tableName, int returnPageID );
	
	// "reactivation" of pageID locks
	// (Note: couldn't put this in StorageManager / StorageManager b/c 
	//        it needs access to the parameter "bit.tableName",
	//		  which is in BufferManager / BufferManager, a level above StorageManager) ..
	// NOTE: this isn't used anymore, because not implementing "timed" locks ..
	void reactivateTablePageLock( int lockNumberIncrement );
	
	
	// getters ..
	map< string, vector<int>* >* getLocks();	// getLocks()->
	
	BufferPool* getBufferPool();			// getBufferPool()->
	
	BufferInfoTable* getBufferInfoTable();  // getBufferInfoTable()->
	
	BITSnapShot* getBitSnapShot();			// getBitSnapShot()->
	
	vector<DataPage*>* getLastFlushedBlocks();		// getLastFlushedBlock()->
	
	vector<int>* getLastFlushedBlockPageIDs();		// getLastFlushedBlockPageID()
	
	
	// setters ..
	void setLocks( map< string, vector<int>* >* newLocks );
	
	void setBufferPool( BufferPool* newBufferPool );
	
	void setBufferInfoTable( BufferInfoTable* newBit );
	
	void setBitSnapShot( BITSnapShot* newBITSnapShot );
	
	void setLastFlushedBlocks( vector<DataPage*>* newLastFlushedBlocks );
	
	void setLastFlushedBlockPageIDs( vector<int>* newLastFlushedBlockPageIDs );
	

private:


	// data members ..
	
	// track when lock "reactivating" is performing ..
	bool lockReactivationOn;
						
	
	// this is necessary in order for "bm.flushBlock()" and "bm.loadBlock()" to call
	// "sm.writeBlock()" and "sm.readBlock()" respectively (wrappers) ..
	StorageManager* sm;
	
	
	// map<tableName, pageID>, tracking the pageID that have active locks in tableName ..
	map< string, vector<int>* >* locks;

	
	// This is the "Buffer Pool" (BP) parameter ..
	// - DataPage* dataPages
	BufferPool* bp;
	
	// table that tracks "BP" metadata ..
	// - string* tableName;
	// - int* bufferPageFieldCount;
	// - int* bufferPageRowCount; // need "bit.bufferPageRowCount" for "bm.findBlockForInsert()" ..
	// - int* recordSize;
	// - int* pageID;
	// - bool* dirty;
	// - int numberOfBlocks;
	BufferInfoTable* bit;
	
	// a "snap shot" of values retrieved from the "bit" in between 
	// operations performed on the BP / bit ..
	BITSnapShot* bitSnapShot;
	
	// keep track of the last block to be flushed ..
	vector<DataPage*>* lastFlushedBlocks;
	
	vector<int>* lastFlushedBlockPageIDs;
	
	static const int blockSize = 48;
	
};

#endif /*BUFFERMANAGER_H_*/
