#include "QueryManager.h"

QueryManager:: QueryManager( int bufferPoolSize, CatalogFile* catalogFile, DatabaseFile* databaseFile ){
	
	//cout << "QueryManager:: QueryManager() start .." << endl << endl;
	
	// QueryManager points to both "StorageManager" and "BufferManager" ..
	// in order that queries that don't involve buffer can be directly performed ..
	//
	//	   qm ->  .. ->  .. -> sm -> cm -> mm
	//		   \  	        /
	//			->   bm   ->
	//           \        /
	//			  -> im ->
	
	MessageManager* mm = new MessageManager();
	
	CatalogManager* cm = new CatalogManager( mm, catalogFile );
	
	
	// QueryManager may want to bypass BufferManager 
	// to access StorageManager directly ..
	sm = new StorageManager( cm, databaseFile );
	
	bm = new BufferManager( bufferPoolSize, sm );
	
	im = new IndexManager( sm );
	
	//cout << "QueryManager:: QueryManager() end .." << endl << endl;
	
}


QueryManager:: ~QueryManager(void){
	
	//cout << "~QueryManager() start .." << endl << endl;

	// "delete" order is vital, b/c "bm" deletes "sm->cm",
	// so must delete "sm" _after_ "bm" ..
	delete bm;
	
	delete sm;
	
	//cout << "~QueryManager() end .." << endl << endl;
	
}

IndexManager* QueryManager:: getIndexManager(){
	
	return im;
	
}

BufferManager* QueryManager:: getBufferManager(){
	
	return bm;
	
}


StorageManager* QueryManager:: getStorageManager(){
	
	return sm;
	
}


CatalogManager* QueryManager:: getCatalogManager(){
	
	return sm->getCatalogManager();
	
}


MessageManager* QueryManager:: getMessageManager(){
	
	return sm->getCatalogManager()->getMessageManager();
	
}


void QueryManager:: createTable( string tableName, vector<string>* attributeNames, int numberOfAttributes ){
	
	if( tableName.length() < 0 )
		return;
		
	if( attributeNames == NULL )
		return;
		
	if( numberOfAttributes < 0 )
		return;
	
	//cout << "QueryManager:: createTable() start .." << endl << endl;
	
	// handles the query "create table xyz(a, b, c)" ..
	
	string returnString;
	
	string errorString;
	
	
	// search the Catalog to see if this "tableName" already exists ..

	 bool doesTableNameExist = getCatalogManager()->doesCatalogTableNameExist( tableName );

	 if ( doesTableNameExist ){
	 	
	 	// 1) the "tableName" already exists in the Catalog,
	 	//    so return an error message ..
	 	
	 	errorString = "\nThe table ";
	 	errorString += tableName;
	 	errorString += " already exists.  The table could not be created [ QM::createTable ].";
	 	
	 	// 2) update the "MessageManager" to track this error and output to user ..
	 	getMessageManager()->setErrorStringQM( errorString );

	 	
	} else {
		
		
	 	// 1) the "tableNameX" does not already exist in the Catalog,
	 	//    so continue creating this table ..

	 	bool createTablePerformed = getStorageManager()->createTable( tableName, numberOfAttributes, attributeNames );
	 	
	 	// if "createTable()" failed, 
	 	// abort and notify the user ..
	 	if ( createTablePerformed == false ) {
	 		
	 		errorString += "\nThe table could not be created [ QM::createTable ].";
	 	
	 		// 2) update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringQM( errorString );
	 		
	 		return;
	 		
	 	}
	 	

	 	returnString = "\nThe table ";
	 	returnString += tableName;
	 	returnString += " was created.";
	 	
	 	// set the resultMessage in the MessageManager ..
	 	getMessageManager()->setResultMessage( returnString );

	 	
	 } // end "if ( doesTableNameExist )" ..
	
	//cout << "QueryManager:: createTable() end .." << endl << endl;
	
	
} // end "createTable()" ..



void QueryManager:: createIndex( string indexName, string tableName, bool isNoDuplicates, string attributeName ){
	
	//cout << "QueryManager:: createIndex() start .." << endl << endl;
	
	// handles the query "create index I on xyz(a) [No Duplicates]" ..
	
	string returnString;
	
	string errorString;
	
	
	// search the Catalog to see if this "indexName" already exists ..
	 bool doesIndexNameExist = getCatalogManager()->doesCatalogIndexNameExist( indexName );

	 //cout << "QueryManager:: createIndex() doesIndexNameExist: " << doesIndexNameExist << endl << endl;
	 
	 if ( doesIndexNameExist ){
	 	
	 	// 1) the "indexName" already exists in the Catalog,
	 	//    so return an error message ..
	 	
	 	errorString = "\nThe index ";
	 	errorString += indexName;
	 	errorString += " already exists.  The index could not be created [ QM::createIndex ].";
	 	
	 	// 2) update the "MessageManager" to track this error and output to user ..
	 	getMessageManager()->setErrorStringQM( errorString );

	 	
	} else {


		// DEBUGGING ..
		//cout << "QM:: index DNE .." << endl << endl;
				
		
	 	// 1) the "indexName" does not already exist in the Catalog,
	 	//    so continue creating this index ..
	 	
	 	bool createIndexPerformed = getIndexManager()->createIndex( indexName, tableName, isNoDuplicates, attributeName );
	 	
	 	//cout << "QueryManager:: createIndex() Point A .." << endl << endl;
	 	
	 	//cout << "QueryManager:: createIndex() createIndexPerformed: " << createIndexPerformed << endl << endl;
	 	 
	 	// if "createIndex()" failed, 
	 	// abort and notify the user ..
	 	if ( createIndexPerformed == false ) {
	 		
	 		errorString += "\nThe index could not be created [ QM::createIndex ].";
	 	
	 		// 2) update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringQM( errorString );
	 		
	 		return;
	 		
	 	}
	 	
		//cout << "QueryManager:: createIndex() Point B .." << endl << endl;

	 	returnString = "\nThe index ";
	 	returnString += indexName;
	 	returnString += " was created.";
	 	
	 	// set the resultMessage in the MessageManager ..
	 	getMessageManager()->setResultMessage( returnString );

	 	
	 } // end "if ( doesIndexNameExist )" ..
	
	//cout << "QueryManager:: createIndex() end .." << endl << endl;
	
	
} // end "createIndex()" ..



void QueryManager:: insertIntoTableValues( string tableNameX, vector<string>* insertValues, int numberOfAttributes ){
	
	//cout << "QueryManager:: insertIntoTableValues() start .." << endl << endl;
	
	if( tableNameX.length() < 0 )
		return;
		
	if( insertValues == NULL )
		return;
		
	if( numberOfAttributes < 0 )
		return;
	
	// handles the query "insert INTO xyz values( a1, a2, .. ax )" ..
	
	//cout << "QueryManager:: insertIntoTableValues() Point A .." << endl << endl;

	if( insertValues == NULL )
		return;

	string returnString;
	
	string errorString;
	
	// 1) ask Catalog if the TABLE-to-insert-into EXISTS ..
	bool doesTableNameXExist = getCatalogManager()->doesCatalogTableNameExist( tableNameX );
	
	
	if ( doesTableNameXExist ){
		
		// (a) ask Catalog if the NUMBER of ATTRIBUTES is correct ..
		int catalogNumberOfAttributes = getCatalogManager()->getNumberOfAttributes( tableNameX );
		
		// (b) number of attributes to be inserted matches the Catalog, so proceed ..
		if ( numberOfAttributes == catalogNumberOfAttributes ){
			
			// insert the values into the table ..
			
			//cout << "QueryManager:: insertIntoTableValues() Point B .." << endl << endl;

			// Pass to "BufferManager" the "insertValues",
			// but converted from string values to int values ..
			vector<int>* insertValuesInt = new vector<int>();
			for(int i=0; i<insertValues->size(); i++){
				

				insertValuesInt->push_back( strToInt( insertValues->at( i ) ) );
				
			}
			
			//cout << "QueryManager:: insertIntoTableValues() Point C .." << endl << endl;
			
			bool isInsertSuccess = getBufferManager()->insertRecordIntoBlock( tableNameX, insertValuesInt, numberOfAttributes );
			
			//cout << "QueryManager:: insertIntoTableValues() Point C3 .." << endl << endl;

			if ( isInsertSuccess ){
				
				returnString = "\nThe record was inserted.";
				
	 			// update the "MessageManager" to track this result and output to user ..
	 			getMessageManager()->setResultMessage( returnString );
				
			} else {
				
				errorString = "\nThe record could not be inserted [ QM::insertIntoXValues ].";
				
	 			// update the "MessageManager" to track this error and output to user ..
	 			getMessageManager()->setErrorStringQM( errorString );
				
			}
			
			// DEBUGGING ..
			//cout << "QueryManager:: insertIntoXValues() Point A .." << endl << endl;
			//cout << "QueryManager:: insertIntoXValues() isInsertSuccess: " << isInsertSuccess << endl << endl;
			//if( isInsertSuccess == true )
			//	cout << "QueryManager:: insertIntoXValues() returnString: " << returnString << endl << endl;
			//else
			//	cout << "QueryManager:: insertIntoXValues() errorString: " << errorString << endl << endl;
			
			// update the Indices associated with this table ..
			vector<string>* tableNames = new vector<string>();
			tableNames->push_back( tableNameX );
			updateIndex( tableNames );
			
			//cout << "QueryManager:: insertIntoTableValues() Point C4 .." << endl << endl;

			
			
		} else {
			
	 		// the "tableNameX" in the Catalog had different number of attributes than attempted with insert,
	 		// so return an error message ..
	 	
	 		errorString = "\nTable \"";
	 		errorString += tableNameX;
	 		errorString += "\" has ";
	 		errorString += intToStr( catalogNumberOfAttributes );
	 		errorString += " attributes. The insert attempted ";
	 		errorString += intToStr( numberOfAttributes );
	 		errorString += " attributes.  The record could not be inserted [ QM::insertIntoXValues ].";
	 		
	 		// update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringQM( errorString );
	 		
	 		
		} // end "if ( numberOfAttributes == catalogNumberOfAttributes )" ..
		
		
	} else {
	 
	 
	 	// the "tableNameX" does not exist in the Catalog,
	 	// so return an error message ..
	 	
	 	errorString = "\nThe table ";
	 	errorString += tableNameX;
	 	errorString += " does not exist.  The record could not be inserted [ QM::insertIntoXValues ].";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
	 	getMessageManager()->setErrorStringQM( errorString );
		
		
	} // end "if ( doesTableNameXExist )" ..
	
	
	//return returnString;
	
	
} // end "insertIntoXValues( string tableNameX, string* attributeNames, int numberOfAttributes )" ..



void QueryManager:: updateIndex( vector<string>* tableNames_ ){
	
	//cout << "QueryManager:: updateIndex() start .." << endl << endl;
	
	vector<DataPage*>* flushedDataPages = getBufferManager()->getLastFlushedBlocks();
	
	//cout << "QueryManager:: updateIndex() Point A1 .." << endl << endl;
	
	vector<int>* flushedPageIDs = getBufferManager()->getLastFlushedBlockPageIDs();
	
	if ( flushedDataPages->size() > 0 ) {
	
		//cout << "QueryManager:: updateIndex() Point A2 .." << endl << endl;
	
		string tableName = "";
		int searchKey = -1;
		int flushedPageID = -1;
	
		//cout << "QueryManager:: updateIndex() Point A3 .." << endl << endl;
	
		bool isIndexUpdated = false;
			
		//vector<DataPageRecord>* flushedDataPageRecords = NULL;
			
		//vector<DataPageField>* flushedDataPageFields = NULL;
	
		//cout << "QueryManager:: updateIndex() Point A4 .." << endl << endl;
	
		int tableNamesSize = tableNames_->size();
		
		vector<string>* indiceNames = NULL;
		
		vector<string>* attributeNames = NULL;
		
		vector< vector<int>* >* flushedEntries = NULL;
		int flushedEntriesSize = -1;
		
		//cout << "QueryManager:: updateIndex() tableNamesSize: " << tableNamesSize << endl << endl;
	
		// loop through each "tableNames" / "flushedDataPages" / "flushedPageID" triple ..
		for(int i=0; i<tableNamesSize; i++){
			
			tableName = tableNames_->at(i);
			
			//cout << "QueryManager:: updateIndex() tableName: " << tableName << endl << endl;
			
			// make sure each "tableName" is non-empty ..
			if( tableName.size() > 0 ){
			
				//flushedDataPageRecords = flushedDataPages->at(i)->getRecords();
				flushedEntries = flushedDataPages->at(i)->getEntries();
				flushedEntriesSize = flushedEntries->size();
				flushedPageID = flushedPageIDs->at(i);
		
				//cout << "QueryManager:: updateIndex() Point B .." << endl << endl;
			
				indiceNames = getCatalogManager()->getCatalogIndexNames( tableName );
				if( indiceNames == NULL )
					indiceNames = new vector<string>();
				
				attributeNames = getCatalogManager()->getCatalogIndexAttributeNames( tableName );
				if( attributeNames == NULL )
					attributeNames = new vector<string>();
			
				//cout << "QueryManager:: updateIndex() Point C .." << endl << endl;
			
				int attributeNamePosition = -1;
				string indexName = "";
				string attributeName = "";
				bool isUpdateIndex = false;
		
				//cout << "QueryManager:: updateIndex() Point D .." << endl << endl;
			
				//cout << "   QueryManager:: updateIndex() indiceNames->size(): " << indiceNames->size() << endl << endl;
				
				//cout << "QueryManager:: updateIndex() attributeNames->size(): " << attributeNames->size() << endl << endl;
			
				//cout << "   QueryManager:: updateIndex() indiceNames == NULL: " << (indiceNames == NULL) << endl << endl;
			
				//cout << "QueryManager:: updateIndex() attributeNames == NULL: " << (attributeNames == NULL) << endl << endl;
			
				bool doesAttributeHaveIndex = false;
				
				int attributesSize = attributeNames->size();
			
				// loop through each "indexName" / "attributeName" pair ..
				for(int j=0; j<attributesSize; j++){
			
					//cout << "QueryManager:: updateIndex() Point E1 .." << endl << endl;
			
					indexName = indiceNames->at(j);
					attributeName = attributeNames->at(j);
			
					//cout << "QueryManager:: updateIndex() Point E2 .." << endl << endl;
					
					doesAttributeHaveIndex = getCatalogManager()->doesCatalogTableAttributeNameHaveIndex( tableName, attributeNames->at(j) );
				
					if( doesAttributeHaveIndex == true ) {
				
						// get the numerical "position" of the attribute .. (0 .. M) ..
						attributeNamePosition = getCatalogManager()->getAttributePosition( tableName, attributeNames->at(j) );

						//cout << "QueryManager:: updateIndex() Point F .." << endl << endl;

						// loop through each "flushedDataPage row" ..
						//for(int k=0; k<flushedDataPageRecords->size(); k++){
						for(int k=0; k<flushedEntriesSize; k++){
				
							// get the "searchKey" from this "attribute" ..
							//flushedDataPageFields = flushedDataPageRecords->at(k).getFields();
							//searchKey = flushedDataPageFields->at( attributeNamePosition ).getValue();
							searchKey = flushedEntries->at(k)->at( attributeNamePosition );
					
							//cout << "QueryManager:: updateIndex() Point G .." << endl << endl;
						
							//cout << "QueryManager:: updateIndex() tableName: " << tableName << endl << endl;
						
							//cout << "QueryManager:: updateIndex() indexName: " << indexName << endl << endl;
						
							//cout << "QueryManager:: updateIndex() attributeName: " << attributeName << endl << endl;
						
							//cout << "QueryManager:: updateIndex() searchKey: " << searchKey << endl << endl;
					
							// update the index ..
							isUpdateIndex = getIndexManager()->updateIndex( indexName, tableName, attributeName, searchKey, flushedPageID );
					
							//cout << "QueryManager:: updateIndex() isIndexUpdated: " << isIndexUpdated << endl << endl;
							//cout << "QueryManager:: updateIndex()  isUpdateIndex: " << isUpdateIndex << endl << endl;
					
							//cout << "QueryManager:: updateIndex() Point H .." << endl << endl;
					
							if( ( isIndexUpdated == false ) && 
								( isUpdateIndex == true ) )
									isIndexUpdated = true;
					
						} // end "for()" ..
						
					} // end "if()" ..
			
				} // end "for()" ..
				
			} // end "if()" ..

		} // end "for()" ..
	
		//cout << "QueryManager:: updateIndex() Point I .." << endl << endl;
		
		//cout << "QueryManager:: updateIndex() isIndexUpdated: " << isIndexUpdated << endl << endl;
	
		// cleanup ..
		getBufferManager()->getLastFlushedBlocks()->clear();
		getBufferManager()->getLastFlushedBlockPageIDs()->clear();
			
		if ( !isIndexUpdated ){
				
			string errorString = "\nThe index could not be updated [ QM::insertIntoXValues ].";
				
	 		// update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringQM( errorString );
				
		}
		
	} // end "if()" ..
	
	//cout << "QueryManager:: updateIndex() Point J .." << endl << endl;
	
} // end "updateIndex()" ..



void QueryManager:: commit(){
	
	// handles the query "commit" ..
	
	//cout << "QueryManager:: commit() Point A .." << endl << endl;
	
	string returnString = "";
	
	string errorString = "";
	//cout << "QueryManager:: commit() Point A1 .." << endl << endl;
	// get the tableNames associated with the buffer pool,
	// get the state prior to calling "commitBufferPool()",
	
	//cout << "QueryManager:: commit() (getBufferManager() == NULL): " << (getBufferManager() == NULL) << endl << endl;
	//cout << "QueryManager:: commit() (getBufferManager()->getBufferInfoTable() == NULL): " << (getBufferManager()->getBufferInfoTable() == NULL) << endl << endl;
	//cout << "QueryManager:: commit() (getBufferManager()->getBufferInfoTable()->getTableName() == NULL): " << (getBufferManager()->getBufferInfoTable()->getTableName() == NULL) << endl << endl;
	
	//vector<string>* bitTableNames = getBufferManager()->getBufferInfoTable()->getTableName();
	
	//cout << "QueryManager:: commit() Point A2 .." << endl << endl;
	
	//cout << "QueryManager:: commit() (bitTableNames == NULL): " << (bitTableNames == NULL) << endl << endl;
	
	//cout << "QueryManager:: commit() bitTableNames: " << (*bitTableNames) << endl << endl;
	
	//int numberOfTableNames = bitTableNames->size();
	
	//int bufferPoolSize = getBufferManager()->getBufferInfoTable()->getBufferPoolSize();
	
	//cout << "QueryManager:: commit() bufferPoolSize: " << bufferPoolSize << endl << endl;
	//cout << "QueryManager:: commit() Point A3 .." << endl << endl;
	vector<string>* tableNames = new vector<string>();
	//for(int i=0; i<bufferPoolSize; i++)
	//	tableNames->push_back( bitTableNames->at(i) );
		//tableNames->push_back( getBufferManager()->getBufferInfoTable()->getTableName()->at(i) );
			
	//cout << "QueryManager:: commit() Point B .." << endl << endl;
	
	// check whether any page in the bp is eligible to be flushed ..
	bool canBufferPoolBeFlushed = getBufferManager()->getBufferInfoTable()->isBufferPoolCommit();
	
	if( canBufferPoolBeFlushed ){
		
		// if at least one page in the bp is eligible to be flushed,
		// proceed with the commit operation ..
		
		bool isCommitSuccess = getBufferManager()->commitBufferPool( tableNames );
	
		//cout << "QueryManager:: commit() Point C .." << endl << endl;
			
		if ( isCommitSuccess ){
				
			returnString = "\nThe pages remaining in the buffer were written to file.";
				
	 		// update the "MessageManager" to track this result and output to user ..
	 		getMessageManager()->setResultMessage( returnString );
				
		} else {
				
			errorString = "\nThe pages remaining in the buffer were NOT written to file [ QM::commit ].";
	 		// update the "MessageManager" to track this error and output to user ..
	 		getMessageManager()->setErrorStringQM( errorString );
				
		}
		
	} else {
		
		// nothing in the buffer pool to "commit",
		// so inform the user that the operation was cancelled ..
		returnString = "\nThe buffer had no eligible pages to commit, so the operation was aborted.";
		
		getMessageManager()->setResultMessage( returnString );	
		
	}

	//cout << "QueryManager:: commit() Point A .." << endl << endl;
	
	// update the Indices associated with this table ..
	updateIndex( tableNames );
	
	//cout << "QueryManager:: commit() Point D .." << endl << endl;

	//return returnString;
	
	
} // end "bool QueryManager:: commit()" ..


DataPage* QueryManager:: getGoodRowsDataPage( DataPage* dataPage, vector< vector<string>* >* queryPairs, vector<bool>* isPairs, string tableNameX ){
	
	if( dataPage == NULL )
		return NULL;
		
	if( queryPairs == NULL )
		return NULL;
		
	if( isPairs == NULL )
		return NULL;
		
	if( tableNameX.size() <= 0 )
		return NULL;
	
	// process the input "dataPage", returning a "dataPage" that has only,
	// the rows that meet the criteria of the query ..
	
	// rows (vector<int>) to return for printing,
	// filled with entries that match in both 
	// "goodRows_namePairs" and "goodRows_nameValuePairs" ..
	vector<int>* goodRows = new vector<int>();
				
	// rows (vector<int>) to return for printing based on "isNamePairs" 
	// section of code ..
	vector<int>* goodRows_namePairs = new vector<int>();
				
	// rows (vector<int>) to return for printing based on "isNameValuePairs"
	// section of code (from "matchingRowsForColumns") ..
	vector<int>* goodRows_nameValuePairs = new vector<int>();
	
	bool isNamePairs = isPairs->at(0);
	
	bool isNameValuePairs = isPairs->at(1);
	
	vector<string>* leftName = queryPairs->at(0);
	
	vector<string>* rightName = queryPairs->at(1);
				
	vector<string>* columnName = queryPairs->at(2);
	
	vector<string>* columnValue = queryPairs->at(3);
	
				
	if ( isNamePairs ) {
					
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() isNamePairs == true found .." << endl << endl;
					
		int leftDPValue = -1;
					
		int rightDPValue = -1;
					
		int leftAttributePosition = -1;
					
		int rightAttributePosition = -1;
					
		int namePairCount = 0;
					
		if( leftName->size() == rightName->size() ) {
						
			// DEBUGGING ..
			//cout << "QM:: getGoodRowsDataPage() Point C1 .. " << endl << endl;
			
			//int dataPageRecordsSize = dataPage->getRecords()->size();
			int dataPageRecordsSize = dataPage->getNumberOfRows();
						
			for(int i=0; i<dataPageRecordsSize; i++){
							
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() Point C2[i:"<<i<<"] .. " << endl << endl;
							
				// rows ..
							
				for(int j=0; j<leftName->size(); j++){
								
					// each column (that is compared) in the row ..
								
					leftAttributePosition = getCatalogManager()->getAttributePosition( tableNameX, leftName->at(j) );
								
					rightAttributePosition = getCatalogManager()->getAttributePosition( tableNameX, rightName->at(j) );
								
					//leftDPValue = dataPage->getRecords()->at( i ).getFields()->at( leftAttributePosition ).getValue();
								
					//rightDPValue = dataPage->getRecords()->at( i ).getFields()->at( rightAttributePosition ).getValue();
					
					leftDPValue = dataPage->getEntries()->at( i )->at( leftAttributePosition );
								
					rightDPValue = dataPage->getEntries()->at( i )->at( rightAttributePosition );
								
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() Point C3[i:"<<i<<"][j:"<<j<<"] .. " << endl << endl;
					//cout << "QM:: getGoodRowsDataPage() leftDPValue("<<leftDPValue<<") == rightDPValue("<<rightDPValue<<"): " << ( leftDPValue == rightDPValue ) << endl << endl;
								
					if ( leftDPValue == rightDPValue ) {
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() Point C4[i:"<<i<<"][j:"<<j<<"] .. " << endl << endl;
									
						// this indicates that the data entries for these two columns are equal ..
						namePairCount++;
									
					}
								
				} // end "for(int j=0; j<leftName.size(); j++)" ..
							
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() leftName.size()("<<leftName.size()<<") == namePairCount("<<namePairCount<<"): " << ( leftName.size() == namePairCount ) << endl << endl;
							
				if ( leftName->size() == namePairCount ) {
								
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() Point C5[i:"<<i<<"] .. " << endl << endl;
								
					// this indicates that all "namePairs" for this row were equal,
					// and the namePairs portion of the "where" clause condition 
					// was met for this row,
					// so go ahead and mark this row for printing ..
					goodRows_namePairs->push_back( i );
								
				}
							
							
				// reset this parameter for the next row iteration ..
				namePairCount = 0;
							
							
			} // end "for(int i=0; i<bpDataPage.records.size(); i++)" ..
						
		} // end "if( leftName.size() == rightName.size() )" ..
					
	} // end "if ( isNamePairs )" ..
				
				
	if ( isNameValuePairs ) {
					
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() isNameValuePairs == true found .." << endl << endl;
					
		// check page for values that meet the criteria of the query..
		//int bitIndex = BufferInfoTable::bitIndex;
				
		int dataPageValue = -1;
					
		int attributePosition = -1;
					
		int currentRowNumber = -1;
					
		// DEBUGGING ..
		int currentColumnNumber = -1;
					
		bool foundNewRow = true;
					
		// rows (vector<...>) of columns (vector<int>) for each "where" clause statement
		// that meet the condition(s) ..
		vector< vector<int>* >* rowsOfColumnsConditionMet = new vector< vector<int>* >();

		// column positions per row ..
		vector<int>* whereClauseColumnPositions = new vector<int>();
					
		// rows that have at least 1 matching column (includes bad rows) ..
		vector<int>* matchingRows = new vector<int>();
					
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() columnName.size()("<<columnName.size()<<") == columnValue.size()("<<columnValue.size()<<"): " << ( columnName.size() == columnValue.size() ) << endl << endl;
					
		int columnNameSize = columnName->size();
					
		if( columnNameSize == columnValue->size() ) {
						
			// DEBUGGING ..
			//cout << "QM:: getGoodRowsDataPage() Point C6 .. " << endl << endl;'
			
			int dataPageNumberOfRows = dataPage->getNumberOfRows();
						
			for(int i=0; i<dataPageNumberOfRows; i++){
							
				// each row ..
							
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() Point C7[i:"<<i<<"] .. " << endl << endl;
							
				for(int j=0; j<columnNameSize; j++){
								
					// each column (that is compared) in the row ..
								
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() Point C8[j:"<<j<<"] .. " << endl << endl;
					//cout << "QM:: getGoodRowsDataPage() leftName[j:"<<j<<"]: " << columnName[j] << endl << endl;
		
					attributePosition = getCatalogManager()->getAttributePosition( tableNameX, columnName->at(j) );
		
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() attributePosition: " << attributePosition << endl << endl;
		
		
					if ( whereClauseColumnPositions->size() < columnNameSize ) {
										
						// fill a container with the relevant "attributePositions",
						// but only do this once per "bpDataPage" ..
						whereClauseColumnPositions->push_back( attributePosition );
									
					}
		
					dataPageValue = dataPage->getEntries()->at( i )->at( attributePosition );
		
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() columnValue[j:"<<j<<"]("<<columnValue[j]<<") == dpValue("<<dpValue<<"): " << ( strToInt( columnValue[j] ) == dpValue ) << endl << endl;
		
					if ( strToInt( columnValue->at(j) ) == dataPageValue ) {
									
						// the value in "bpDataPage" matches the "where" query value ..
									
						if ( foundNewRow ) {
										
							// only add "*new vector<int>()" for new rows ..
										
							currentRowNumber++;
										
							// DEBUGGING ..
							//cout << "QM:: getGoodRowsDataPage() currentRowNumber: " << currentRowNumber << endl << endl;
							//cout << "QM:: getGoodRowsDataPage() new vector<int>() pushed .." << endl << endl;

							rowsOfColumnsConditionMet->push_back( new vector<int>() );
										
							foundNewRow = false;
										
							// for each "rowsOfColumnsConditionMet" entry,
							// track the row that it is in (used far below to find "goodRows") ..
							matchingRows->push_back( i );
										
							// DEBUGGING ..
							//cout << "QM:: getGoodRowsDataPage() matchingRows.push_back( "<<i<<" ) .." << endl << endl;
										
						}
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() Point C9[j:"<<j<<"] .. " << endl << endl;
									
						// "where" clause condition is met for this row[ currentRowNumber ] / this column[j] ..
						rowsOfColumnsConditionMet->at(currentRowNumber)->push_back( attributePosition );	
										
						// DEBUGGING ..
						currentColumnNumber++;
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() currentRowNumber: " << currentRowNumber << endl << endl;
						//cout << "QM:: getGoodRowsDataPage() currentColumnNumber: " << currentColumnNumber << endl << endl;
						//cout << "QM:: getGoodRowsDataPage() matchingColumnsForRows["<<currentRowNumber<<"].push_back("<<attributePosition<<"): " << rowsOfColumnsConditionMet[ currentRowNumber ][ currentColumnNumber ] << endl << endl;
								

					} else {
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() Point C10[j:"<<j<<"] .. " << endl << endl;
									
						// "where" clause condition not met ..
						continue;
									
					} // end "if ( rightValue[j] == dpValue )" ..
		
		
				} // end "for(int j=0; j<columnName.size(); j++)" ..
							
							
				// reset this parameter for the next row iteration ..
				currentColumnNumber = -1;
				foundNewRow = true;
						
						
			} // end "for(int i=0; i<dataPage.records.size(); i++)" .. 
						
						
			int matchingColumnCount = 0;
			
			int whereClauseColumnPositionsSize = whereClauseColumnPositions->size();
			
			int rowsOfColumnsConditionMetSize = rowsOfColumnsConditionMet->size();
						
			for(int i=0; i<rowsOfColumnsConditionMetSize; i++){
							
				// rows ..
							
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() Point C11[i:"<<i<<"] .. " << endl << endl;
							
				if ( rowsOfColumnsConditionMet->at(i)->size() == whereClauseColumnPositionsSize ) {
								
					for(int j=0; j<whereClauseColumnPositionsSize; j++){
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() whereClauseColumnPositions[j:"<<j<<"]("<<whereClauseColumnPositions[j]<<") == rowsOfColumnsConditionMet[i:"<<i<<"][j:"<<j<<"]("<<rowsOfColumnsConditionMet[i][j]<<"): " << ( whereClauseColumnPositions[j] == rowsOfColumnsConditionMet[i][j] ) << endl << endl;
									
						if ( whereClauseColumnPositions->at(j) == rowsOfColumnsConditionMet->at(i)->at(j) ){
											
							// if "rowsOfColumnsConditionMet" value found in "whereClauseColumnPositions",
							// then so far this row is ok to return for print ..
										
							// DEBUGGING ..
							//cout << "QM:: getGoodRowsDataPage() Point C12[i:"<<i<<"][j:"<<j<<"] .. " << endl << endl;
										
							matchingColumnCount++;
										
							// DEBUGGING ..
							//cout << "QM:: getGoodRowsDataPage() matchingColumnCount: " << matchingColumnCount << endl << endl;
											
						} // end "if ( whereClauseColumnPositions[j] == matchingColumnsForRows[i][j] )" ..
									
									
					} // end "for(int j=0; j<whereClauseColumnPositions.size(); j++)" ..
								

					if( matchingColumnCount == whereClauseColumnPositionsSize ){
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() Point C13[i:"<<i<<"] .. " << endl << endl;
									
						// track this row in "goodRows" as a row to print ..
						goodRows_nameValuePairs->push_back( matchingRows->at(i) );
									
					}
								
								
				} // end "if ( matchingColumnsForRows[i].size() == whereClauseColumnPositions.size() )" ..
								
							
				// update this parameter for the next row iteration ..
				matchingColumnCount = 0;


			} // end "for(int i=0; i<matchingColumnsForRows.size(); i++)" ..
						
						
						
		} // end "if( leftName.size() == rightValue.size() )" ..
					
				
	} // end "if ( isNamePairs )" ..
				
				
	// rectify "goodRows_namePairs" and "goodRows_nameValuePairs" 
	// to fill up "goodRows" ..
				
	if( isNamePairs ) {
					
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() Point C14 .. " << endl << endl;
					
		// ( isNamePairs == true ) && 
		// ( isNameValuePairs == true || false ), so
		// only use rows that appear in [both] "goodRows_nameValuePairs" and
		// "goodRows_namePairs", and put those rows into "goodRows" ..
					
		if ( goodRows_nameValuePairs->size() >= goodRows_namePairs->size() ){
						
			// ( isNamePairs == true ) && 
			// ( isNameValuePairs == true ) ..
						
			//  MORE: "goodRows_nameValuePairs" ..
			// FEWER: "goodRows_namePairs" ..
						
			// NOTE: b/c "isNamePairs" means "goodRows_namePairs.size() > 0",
			// this also means "goodRows_nameValuePairs.size() >= 0" here
			// due to above condition ..
						
			// DEBUGGING ..
			//cout << "QM:: getGoodRowsDataPage() Point C15 .. " << endl << endl;
						
			for(int i=0; i<goodRows_namePairs->size();i++){
							
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() Point C16 .. " << endl << endl;
						
				for(int j=0; j<goodRows_nameValuePairs->size();j++){
								
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() Point C17 .. " << endl << endl;
								
					// check each "goodRows_namePairs" against all of
					// the "goodRows_nameValuePairs" ..
					
					if( goodRows_nameValuePairs->at(j) == goodRows_namePairs->at(i) ){
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() Point C18 .. " << endl << endl;
								
						// "goodRows_namePairs" row is ok ..
						goodRows->push_back( goodRows_namePairs->at(i) );
						break;
								
					}
					
					
				} // end "for(int j=0; j<goodRows_nameValuePairs.size();j++)" ..
						
						
			} // end "for(int i=0; i<goodRows_namePairs.size();i++)" ..
						
						
		} else if ( ( goodRows_namePairs->size() > goodRows_nameValuePairs->size() ) &&
				    ( goodRows_nameValuePairs->size() > 0 ) ) {
						     	
			// DEBUGGING ..
			//cout << "QM:: getGoodRowsDataPage() Point C19 .. " << endl << endl;
							
			// ( isNamePairs == true ) && 
			// ( isNameValuePairs == true ) ..
						
			//  MORE: "goodRows_namePairs" ..
			// FEWER: "goodRows_nameValuePairs" ..
						
			// NOTE: "isNamePairs" means "goodRows_namePairs.size() > 0",
			// and above states that "goodRows_nameValuePairs.size() > 0" 
			// also, this doesn't handle when "goodRows_nameValuePairs.size() == 0"
			// (see next if() condition for that) ..
						
			for(int i=0; i<goodRows_nameValuePairs->size();i++){
							
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() Point C20 .. " << endl << endl;
						
				for(int j=0; j<goodRows_namePairs->size();j++){
								
					// DEBUGGING ..
					//cout << "QM:: getGoodRowsDataPage() Point C21 .. " << endl << endl;
								
					// check each "goodRows_nameValuePairs" against all of
					// the "goodRows_namePairs" ..
					
					if( goodRows_namePairs->at(j) == goodRows_nameValuePairs->at(i) ){
									
						// DEBUGGING ..
						//cout << "QM:: getGoodRowsDataPage() Point C22 .. " << endl << endl;
								
						// "goodRows_nameValuePairs" row is ok ..
						goodRows->push_back( goodRows_nameValuePairs->at(i) );
						break;
								
					}
					
					
				} // end "for(int j=0; j<goodRows_namePairs.size();j++)" ..
						 
						 
			} // end "for(int i=0; i<goodRows_nameValuePairs.size();i++)" ..
							
							
		} else if ( goodRows_nameValuePairs->size() == 0 ) {

			// ( isNamePairs == true ) && 
			// ( isNameValuePairs == true || false ) ..
						
			// NOTE: b/c "isNamePairs" means "goodRows_namePairs.size() > 0",
			// so only check for "goodRows_nameValuePairs.size() == 0" ..
						
			// DEBUGGING ..
			//cout << "QM:: getGoodRowsDataPage() Point C23 .. " << endl << endl;
						
			// if "isNameValuePairs == true" && 
			// "goodRows_nameValuePairs.size() == 0": don't do below,
			// so only do this when: "isNameValuePairs == false" ..
						
			if ( isNameValuePairs == false ) {
							
				// just transfer "goodRows_namePairs" values into
				// "goodRows" ..
							
				for(int i=0; i<goodRows_namePairs->size(); i++){
						
					goodRows->push_back( goodRows_namePairs->at(i) );
						
				}
							
			}
					
		} // end "if ( ( goodRows_nameValuePairs.size() == 0 )" ..
					

					
	} else if ( isNameValuePairs ) {
					
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() Point C24 .. " << endl << endl;
					
		// ( isNamePairs == false ) && 
		// ( isNameValuePairs == true ), so
		// just transfer "goodRows_nameValuePairs" values into
		// "goodRows" ..
					
		for(int i=0; i<goodRows_nameValuePairs->size(); i++){
						
			goodRows->push_back( goodRows_nameValuePairs->at(i) );
						
		}
					
	} // end "if( isNamePairs )" ..

				
	// DEBUGGING ..
	//cout << "QM:: getGoodRowsDataPage() Point C25 .. " << endl << endl;
						
	int goodRow = -1;
				
	//int rowCount = 0;
						
	DataPage* goodBPDataPage = new DataPage();
				
	int goodRowsSize = goodRows->size();
	
	int dataPageNumberOfRows = dataPage->getNumberOfRows();
				
	// get the "goodRows" out of the bpDataPage for display ..
	for(int i=0; i<goodRowsSize; i++){
								
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() Point C26[i:"<<i<<"] .. " << endl << endl;
						
		goodRow = goodRows->at(i);
								
		// DEBUGGING ..
		//cout << "QM:: getGoodRowsDataPage() goodRow: " << goodRow << endl << endl;
						
		//goodBPDataPage->getRecords()->push_back( ( *new DataPageRecord() ) );
							
		//for(int j=0; j<dataPage->getRecords()->at( 0 ).getFields()->size(); j++){
						
			// DEBUGGING ..
			//cout << "QM:: getGoodRowsDataPage() Point C27[i:"<<i<<"][j:"<<j<<"] .. " << endl << endl;
						
		if ( goodRow < dataPageNumberOfRows ) {
										
				// DEBUGGING ..
				//cout << "QM:: getGoodRowsDataPage() Point C28[i:"<<i<<"][j:"<<j<<"] .. " << endl << endl;
								
				//goodBPDataPage->getRecords()->at( rowCount ).getFields()->push_back( ( *new DataPageField() ) );
						
				//goodBPDataPage->getRecords()->at( rowCount ).getFields()->at( j ).setValue( dataPage->getRecords()->at( goodRow ).getFields()->at( j ).getValue() );
						
			goodBPDataPage->addEntries( dataPage->getEntries( goodRow ) );
								
		} // end "if()" ..
						
		//} // end "for()" ..
						
						
		// update this parameter for the next iteration of the next good row ..
		//rowCount++;
						
						
	} // end "for()" ..
						
						
	// update "dataPage" to include only the "good" rows ..
	// (used in "selectFromX()" for "bpReturnPages") ..
	return goodBPDataPage;	
	
	
} // end "getGoodRowsDataPage( DataPage dataPage, vector< vector<string> > queryPairs )" ..


map<int, DataPage* >* QueryManager:: selectBuffer( string tableNameX ) {
	
	// handles the query "select * from X", 
	// when returning results from the buffer pool ..
	
	//cout << "QM:: printBuffer() start .." << endl << endl;

	DataPage* bpDataPage = new DataPage();
	
	map<int, DataPage* >* bpReturnPages = new map<int, DataPage* >();
	
	string returnString = "";
	
	string errorString = "";
	
	if( tableNameX.length() <= 0 )
		return bpReturnPages;
	
	
	// ask Catalog if the TABLE-to-select-from EXISTS ..
	bool doesTableNameXExist = getCatalogManager()->doesCatalogTableNameExist( tableNameX );
	
	if ( doesTableNameXExist ){
		
		// 1) take a "snap shot" of values from the "bit" ( pageID, bufferPoolSize, LRUBlocks, rowAvailable )
		//    prior to operations being performed on the BP / "bit" values changing ..
		
		// bitSnapShop.tableName: track the snap shot of tableName in "bit" ..
		getBufferManager()->getBitSnapShot()->assignTableNames( getBufferManager()->getBufferInfoTable()->getTableName() );
		
		// bitSnapShop.pageID: track the snap shot of pageID in "bit" ..
		getBufferManager()->getBitSnapShot()->assignPageID( getBufferManager()->getBufferInfoTable()->getPageID() );
		
		// bitSnapShop.bufferPoolSize: track the snap shot of bufferPoolSize in "bit" ..
		getBufferManager()->getBitSnapShot()->setBufferPoolSize( getBufferManager()->getBufferInfoTable()->getBufferPoolSize() );
		
		
		// 2) bitSnapShot.LRUBlocks: track the LRU bit blocks with pageID == -1 for this tableName ..
		list<int>::iterator it_LRU;
		
		getBufferManager()->getBitSnapShot()->getLRUBlocks()->clear();

		for(it_LRU = getBufferManager()->getBufferInfoTable()->getLRUBlocks()->begin(); it_LRU != getBufferManager()->getBufferInfoTable()->getLRUBlocks()->end(); it_LRU++){

			// DEBUGGING ..
			//cout << "QM:: selectStarFromX() (*it_LRU): " << (*it_LRU) << endl << endl;

			if ( ( getBufferManager()->getBufferInfoTable()->getTableName()->at(*it_LRU).compare( tableNameX ) == 0 ) &&
			     ( getBufferManager()->getBufferInfoTable()->getPageID()->at(*it_LRU) < 0 ) ) {


				// DEBUGGING ..
				//cout << "QM:: selectStarFromX() bitSnapShot.LRUBlocks.push_back( "<<(*it_LRU)<<" ) )" << endl << endl;
				
				getBufferManager()->getBitSnapShot()->getLRUBlocks()->push_back( (*it_LRU) );
				
			}
			
		} // end "for()" ..
		
			
		// 3) bitSnapShot.rowAvailable: track the snap shot of "rowAvailable" in the "bit" ..
		getBufferManager()->getBitSnapShot()->assignSlotAvailable( getBufferManager()->getBufferInfoTable()->getRowAvailable() );

		
		// Either pass to "BufferManager" data that was
		// either (a) select from existing BP block,
		// or (b) read a page from the file into a BP block and select from it ..
		
		// get a collection of all the "pageID"s 
		// belonging to this table ..

		// check for pages from the BP ..
		int bufferPoolSize = getBufferManager()->getBufferPool()->getBufferPoolSize();
		for( int i=0; i<bufferPoolSize; i++ ) {
			
			if ( ( !getBufferManager()->getBufferInfoTable()->getEmpty()->at(i) ) &&
			     ( getBufferManager()->getBufferInfoTable()->getTableName()->at(i).compare( tableNameX ) == 0 ) ) {
				
				// from the BP, collect all the "pageID"s for this "tableName",
				// use to collect all the "returnDataPage"s ..
				//returnDataPages.push_back( bm->returnBlock( tableNameX, bm->bit.pageID[i] ) );
				//bpDataPages.push_back( bm->returnBPBlock( tableNameX, i ) );
				
				bpDataPage = getBufferManager()->returnBPBlock( i );
				
				//if ( bpDataPage->getRecords()->size() > 0 ) {
				if ( bpDataPage->getNumberOfRows() > 0 ) {
					
					if ( getBufferManager()->getBufferInfoTable()->getPageID()->at(i) < 0 ) {
					
						// DEBUGGING ..
						//cout << "QM:: selectStarFromX() Point C29 .. " << endl << endl;
					
						// since BP pages that have yet to be written to file do not yet
						// have a pageID, where "bm->bit.pageID[i] == -1",
						// and since the "map" container "bpReturnPages" cannot duplicate keys == -1,
						// instead use "bpOffset" as the key for the "bpDataPage".
						// i.e.: i = 0, bpOffset = -1
						//		 i = 1, bpOffset = -2, etc ..
					
						int bpOffset = ( (-1) * (i+1) );
						
						bpReturnPages->insert( pair< int, DataPage* >( bpOffset, bpDataPage ) );
					
					} else {

						bpReturnPages->insert( pair< int, DataPage* >( getBufferManager()->getBufferInfoTable()->getPageID()->at(i), bpDataPage ) );	


					} // end "if ( getBufferManager()->bit.pageID[i] < 0 )" ..
					
					
				} // end "if ( bpDataPage.records.size() > 0 )" ..
				
									
			} // end "if ( ( !getBufferManager()->bit.empty[ i ] ) && " ..
			
			
		} // end "for()" ..
		
		

	} else {
	 
	 	// the "tableNameX" does not exist in the Catalog,
	 	// so return an error message ..
	 	
	 	errorString = "\nThe table ";
	 	errorString += tableNameX;
	 	errorString += " does not exist [ QM::selectBuffer ].";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );	// string detailing the error in "QueryManager" ..
		
		
	} // end "if ( doesTableNameXExist )" ..
	
	
	return bpReturnPages;
	
	
} // end "selectBuffer()" ..


DataPage* QueryManager:: processDataForWhereClause( DataForWhereClauseAggregation* dfwcp ){

	if( dfwcp == NULL )
		return NULL;

	// get out the parameters from the objects ..
	DataPage* dataPage = dfwcp->getDataPage();
	vector< vector<string>* >* queryPairs = dfwcp->getQueryPairs();
	vector<bool>* isPairs = dfwcp->getIsPairs();
	string tableNameX = dfwcp->getTableNameX();
	bool doesQueryHaveWhereClause = dfwcp->getDoesQueryHaveWhereClause();
	bool querySpecifiesColumnNames = dfwcp->getQuerySpecifiesColumnNames();
	map<int, string>* columnNames = dfwcp->getColumnNames();
	int maxNumberOfAttributes = dfwcp->getMaxNumberOfAttributes();
	
	if( dataPage == NULL )
		return NULL;
		
	if( queryPairs == NULL )
		return NULL;
		
	if( isPairs == NULL )
		return NULL;
		
	if( tableNameX.length() < 0 )
		return NULL;
		
	if( dfwcp == NULL )
		return NULL;
		
	if( maxNumberOfAttributes < 0 )
		return NULL;

	// 1) process "fileDataPage", removing rows that don't meet "where" clause criteria ..
	if ( doesQueryHaveWhereClause ) {
					
		dataPage = getGoodRowsDataPage( dataPage, queryPairs, isPairs, tableNameX );	
					
	}
	
	// DEBUGGING ..
	//int columnNamesSize = -1;
	//if( columnNames != NULL )
	//	columnNamesSize = columnNames->size();
	//for(int i=0; i<columnNamesSize; i++)
	//	cout << "QueryManager:: processDataForWhereClause() columnNames->at(i("<<i<<")): " << (*columnNames)[i] << endl << endl;		

	// 2) process "fileDataPage", removing rows that don't meet "where" clause criteria ..
	if ( querySpecifiesColumnNames ) {
		
		DataPage* dataPageReduced = new DataPage();
					
		// (a) the number of name/value pairs should be less than or equal to the max num of attributes ..
		if ( columnNames->size() <= maxNumberOfAttributes ) {
				
			// (i) need to keep only the columns that were indicated in the query,
			//     so leave out the columns from the "fileDataPage" that weren't listed in the query ..
					
			int numberOfRowsInDataPage = dataPage->getNumberOfRows();
					
			//for(int j=0; j<numberOfRowsInDataPage; j++){
						
			//	dataPageReduced->getRecords()->push_back( ( *new DataPageRecord() ) );
						
			//}
			
			int fieldNumber = -1;
			
			vector<int>* entries = NULL;
			
			int numberOfRows = dataPage->getNumberOfRows();
			
			for(int i=0; i<numberOfRows; i++){
				
				entries = dataPage->getEntries( i );
				
				dataPageReduced->toggleNewRow();
				
				for(int j=0; j<maxNumberOfAttributes; j++){
						
					if( columnNames->find( j ) != columnNames->end() ){
					
						dataPageReduced->addEntry( entries->at( j ) );
					
					}
				
				}
				
			}
					
/*		
			int fieldNumber = -1;
					
			for(int j=0; j<maxNumberOfAttributes; j++){
						
				if( columnNames->find( j ) != columnNames->end() ){
							
					//columnNames.find( j )->second;
							
					fieldNumber++;
							
					for(int k=0; k<numberOfRowsInFileDataPage; k++){
		
						dataPageReduced->getRecords()->at( k ).getFields()->push_back( ( *new DataPageField() ) ); 
		
						// copy the BP "dataPage" into the "returnDataPage" ..
						dataPageReduced->getRecords()->at( k ).getFields()->at( fieldNumber ).setValue( dataPage->getRecords()->at( k ).getFields()->at( j ).getValue() );
		
					} // end "for()" ..
							
				} // end "if( columnNames.find( j ) != columnNames.end() )" ..
						
			} // end "for()" ..
*/
					
			// set "fileDataPage" to the "reduced column version" for the section below ..
			dataPage = dataPageReduced;
					
					
		} // end "if ( columnNames.size() < maxNumberOfAttributes )" ..	
					
	} // end "if ( querySpecifiesColumnNames )" ..
	
	
	return dataPage;
	
	
} // end "" ..


vector< map<int, DataPage* >* >* QueryManager:: selectFromX( string tableNameX, vector<string>* selectColumnNames, multimap<string, string>* whereColumnNamePairs, map<string, int>* whereColumnNameValuePairs ) {
	
	// handles the query "select * from X" ..
	
	//cout << "QueryManager:: selectFromX() start .." << endl << endl;
	
	if( tableNameX.size() <= 0 )
		return NULL;
	
	if( selectColumnNames == NULL )
		return NULL;

	if( whereColumnNamePairs == NULL )
		return NULL;
		
	if( whereColumnNameValuePairs == NULL )
		return NULL;
	
	//cout << "QueryManager:: selectFromX() tableNameX: " << tableNameX << endl << endl;
	
	// use this as a flag within "selectFromX()" to toggle the use of
	// method calls that process the "where" clause portion of queries ..
	// (if the query has no "where" clause, this flag should toggle the
	//  method call off) ..
	bool doesQueryHaveWhereClause = false;
	
	// use this flag to toggle on/off whether or not to display only
	// certain columns of the data ("querySpecifiesColumnNames = true",
	// "select a, bb from abc"), or to display all relevant columns of data
	// ("querySpecifiesColumnNames = false", "select * from abc") ..
	bool querySpecifiesColumnNames = false;
	
	if ( selectColumnNames->size() > 0 ) {
		
		querySpecifiesColumnNames = true;
		
	}
	
	//cout << "QueryManager:: selectFromX() Point A1 .." << endl << endl;
	
	// used for (aa == bb) ..
	vector<string>* leftName = new vector<string>();
	
	vector<string>* rightName = new vector<string>();
	
	
	// used for (aa == 25) ..
	vector<string>* columnName = new vector<string>();
	
	vector<int>* columnValue = new vector<int>();
	
	//cout << "QueryManager:: selectFromX() Point A2.." << endl << endl;
	
	
	bool isNamePairs = false;
	
	bool isNameValuePairs = false;
	
	if ( whereColumnNamePairs->size() > 0 ) {
		
		multimap<string, string>::iterator it_names;
		for(it_names = whereColumnNamePairs->begin(); it_names != whereColumnNamePairs->end(); it_names++){
		
			leftName->push_back( it_names->first );
			
			rightName->push_back( it_names->second );
		
			// DEBUGGING ..
			//cout << "QueryManager:: selectFromX() whereColumn[Names]->first: " << it_names->first << endl << endl;
			//cout << "QueryManager:: selectFromX() whereColumn[Names]->second: " << it_names->second << endl << endl;
			//cout << "--------------------------------------------------" << endl << endl;
		
		}
		
		if ( doesQueryHaveWhereClause == false ) {
			
			doesQueryHaveWhereClause = true;
			
		}
		
		isNamePairs = true;
		
	}
	
	//cout << "QueryManager:: selectFromX() Point B.." << endl << endl;
	
	if ( whereColumnNameValuePairs->size() > 0 ) {
		
		map<string, int>::iterator it_nameValues;
		for(it_nameValues = whereColumnNameValuePairs->begin(); it_nameValues != whereColumnNameValuePairs->end(); it_nameValues++){
		
			columnName->push_back( it_nameValues->first );
			
			columnValue->push_back( it_nameValues->second );
		
			// DEBUGGING ..
			//cout << "QueryManager:: selectFromX() whereColumnName[Values]->first: " << it_nameValues->first << endl << endl;
			//cout << "QueryManager:: selectFromX() whereColumnName[Values]->second: " << it_nameValues->second << endl << endl;
			//cout << "--------------------------------------------------" << endl << endl;
		
		}
		
		if ( doesQueryHaveWhereClause == false ) {
			
			doesQueryHaveWhereClause = true;
			
		}
		
		isNameValuePairs = true;
		
	}
	
	//cout << "QueryManager:: selectFromX() Point C.." << endl << endl;
	
	// DEBUGGING ..
	//vector<string> whereColumnNames = whereColumnPairs[0];
	//vector<string> whereColumnNameValues = whereColumnPairs[1];
	//for(int i=0; i<whereColumnNames.size(); i++){
		//cout << "QueryManager:: selectStarFromX() whereColumnNames: " << whereColumnNames[i] << endl << endl;
		//cout << "QueryManager:: selectStarFromX() whereColumnNameValues: " << whereColumnNameValues[i] << endl << endl;
	//}


	DataPage* bpDataPage = new DataPage();
	
	map<int, DataPage* >* bpReturnPages = new map<int, DataPage* >();
	

	DataPage* fileDataPage = new DataPage();
	
	map<int, DataPage* >* fileReturnPages = new map<int, DataPage* >();
	
	string returnString = "";
	
	string errorString = "";
	
	//cout << "QueryManager:: selectFromX() Point D.." << endl << endl;
	
	// DEBUGGING ..
	//cout << "QueryManager:: tableNameX: " << tableNameX << endl << endl;
	
	// ask Catalog if the TABLE-to-select-from EXISTS ..
	bool doesTableNameXExist = getCatalogManager()->doesCatalogTableNameExist( tableNameX );
	
	if ( doesTableNameXExist ){
		
		//cout << "QueryManager:: selectFromX() Point E.." << endl << endl;
		
		// 1) match valid column names with their position order (from left to right)
		//    as it is recorded in the catalog ..
		//    (use for both "bpDataPage" and "fileDataPage") ..
		map<int, string>* columnNames = groupSelectColumnNames( tableNameX, selectColumnNames );
		
		if( columnNames == NULL )
			return NULL;
		
		//cout << "QueryManager:: selectFromX() Point E1.." << endl << endl;
		
		//int maxNumberOfAttributes = bm->cm->getNumberOfAttributes( tableNameX );
		int maxNumberOfAttributes = getCatalogManager()->getNumberOfAttributes( tableNameX );
		
		//cout << "QueryManager:: selectFromX() Point E2.." << endl << endl;
		
		
		// 2) take a "snap shot" of values from the "bit" ( pageID, bufferPoolSize, LRUBlocks, rowAvailable ) prior
		//    to operations being performed on the BP / "bit" values changing ..
		
		// bitSnapShot.tableName: track the snap shot of tableName in "bit" ..
		getBufferManager()->getBitSnapShot()->assignTableNames( getBufferManager()->getBufferInfoTable()->getTableName() );
		
		//cout << "QueryManager:: selectFromX() Point E3.." << endl << endl;
		
		// bitSnapShot.pageID: track the snap shot of pageID in "bit" ..
		getBufferManager()->getBitSnapShot()->assignPageID( getBufferManager()->getBufferInfoTable()->getPageID() );
		
		//cout << "QueryManager:: selectFromX() Point E4.." << endl << endl;
		
		// bitSnapShot.bufferPoolSize: track the snap shot of bufferPoolSize in "bit" ..
		getBufferManager()->getBitSnapShot()->setBufferPoolSize( getBufferManager()->getBufferInfoTable()->getBufferPoolSize() );
		
		//cout << "QueryManager:: selectFromX() Point E5.." << endl << endl;
		
		// 3) bitSnapShot.LRUBlocks: track the LRU bit blocks with pageID == -1 for this tableName ..
		list<int>::iterator it_LRU;
		
		//cout << "QueryManager:: selectFromX() Point F.." << endl << endl;
		

		getBufferManager()->getBitSnapShot()->getLRUBlocks()->clear();

		for(it_LRU = getBufferManager()->getBufferInfoTable()->getLRUBlocks()->begin(); it_LRU != getBufferManager()->getBufferInfoTable()->getLRUBlocks()->end(); it_LRU++){
			
			if ( ( getBufferManager()->getBufferInfoTable()->getTableName()->at(*it_LRU).compare( tableNameX ) == 0 ) &&
			     ( getBufferManager()->getBufferInfoTable()->getPageID()->at(*it_LRU) < 0 ) ) {

				// DEBUGGING ..
				//cout << "QM:: selectFromX() bitSnapShot.LRUBlocks.push_back( "<<(*it_LRU)<<" ) )" << endl << endl;

				getBufferManager()->getBitSnapShot()->getLRUBlocks()->push_back( (*it_LRU) );
				
			}
			
			
		} // end "for()" ..
		
		//cout << "QueryManager:: selectFromX() Point G.." << endl << endl;
		
			
		// bitSnapShot.rowAvailable: track the snap shot of "rowAvailable" in the "bit" ..
		getBufferManager()->getBitSnapShot()->assignSlotAvailable( getBufferManager()->getBufferInfoTable()->getRowAvailable() );

		
		// NOTE: Pass to "BufferManager" data either,
		// (a) select from existing BP block, or
		// (b) read a page from the file into a BP block and select from it ..
		
		
		// 4) update the DataHeader for the query "tableNameX" ..
		DataHeader* header = new DataHeader();
		bool readHeaderPerformed = getStorageManager()->readHeader( tableNameX, header );
		
		// if "readHeader()" failed, 
		// abort "selectFromX()" ..
		if ( readHeaderPerformed == false ) {
				
	 		string errorString = "\nHeader file failed to open[ QM::selectFromX() ], data results may be less than expected.";
	 	
	 		// update the "MessageManager" to track this error and output to user ..
			getMessageManager()->setErrorStringQM( errorString );			
				
			vector< map<int, DataPage* >* >* emptyResults = new vector< map<int, DataPage* >* >();
			
			emptyResults->push_back( bpReturnPages );
			
			emptyResults->push_back( fileReturnPages );
				
			return emptyResults;
				
		}
		
		//cout << "QueryManager:: selectFromX() Point H.." << endl << endl;
		
		bool fileHasPages = false;
		
		if( header->getTotalNumberOfDataPages() > 0 ){
			
			fileHasPages = true;
			
		}
		
		//cout << "QueryManager:: selectFromX() fileHasPages: " << fileHasPages << endl << endl;
		
		
		// 5) get ready for "getGoodRowsDataPage()" ..
		
		vector< vector<string>* >* queryPairs = new vector< vector<string>* >();
				
		queryPairs->push_back( new vector<string>() );
				
		for(int i=0; i<leftName->size(); i++){
					
			queryPairs->at(0)->push_back( leftName->at(i) );
					
		}
				
		queryPairs->push_back( new vector<string>() );
				
		for(int i=0; i<rightName->size(); i++){
					
			queryPairs->at(1)->push_back( rightName->at(i) );
					
		}
				
		queryPairs->push_back( new vector<string>() );
				
		for(int i=0; i<columnName->size(); i++){
					
			queryPairs->at(2)->push_back( columnName->at(i) );
					
		}
				
		queryPairs->push_back( new vector<string>() );
				
		for(int i=0; i<columnValue->size(); i++){
					
			queryPairs->at(3)->push_back( intToStr( columnValue->at(i) ) );
					
		}
		
		vector<bool>* isPairs = new vector<bool>();
		
		isPairs->push_back( isNamePairs );
		
		isPairs->push_back( isNameValuePairs );
		
		//cout << "QueryManager:: selectFromX() Point J.." << endl << endl;
		
		
		vector<int>* bitIndicesForNonEmptyBPBlocks = new vector<int>();;
		
		bool bpBlocksFound = false;
		
		DataForWhereClauseAggregation* dfwcp = new DataForWhereClauseAggregation();
		
		int bufferPoolSize = getBufferManager()->getBufferPool()->getBufferPoolSize();
		
		// 6) check for pages from the BP ..
		for( int i=0; i<bufferPoolSize; i++ ) {
			
			if ( ( !getBufferManager()->getBufferInfoTable()->getEmpty()->at(i) ) &&
			     ( getBufferManager()->getBufferInfoTable()->getTableName()->at(i).compare( tableNameX ) == 0 ) ) {
				
				// (a) from the BP, collect all the "pageID"s for this "tableName",
				//     use to collect all the "returnDataPage"s ..
				
				bpDataPage = getBufferManager()->returnBPBlock( i );
				
				if( bpDataPage == NULL ){
						
					continue;
						
				}
				
				// if "returnBPBlock( i )" fails to return any bp blocks,
				// continue checking the remaining blocks ..
				if ( bpDataPage->getNumberOfRows() == 0 ) {
				
					continue;
				
				}
				
				// need to set all parameters (not just "dataPage"),
				// because may have only "fileDataPage" or just "bpDataPage",
				// (if you had both, these would be set here for "bpDataPage",
				//  and reused below for "fileDataPage" .. but not good enough,
				//  so set them again below for "fileDataPage") ..
				dfwcp->setDataPage( bpDataPage );
				dfwcp->setQueryPairs( queryPairs );
				dfwcp->setIsPairs( isPairs );
				dfwcp->setTableNameX( tableNameX );
				dfwcp->setDoesQueryHaveWhereClause( doesQueryHaveWhereClause );
				dfwcp->setQuerySpecifiesColumnNames( querySpecifiesColumnNames );
				dfwcp->setColumnNames( columnNames );
				dfwcp->setMaxNumberOfAttributes( maxNumberOfAttributes );
				
				// (b) process data according to "where" clause criteria ..
				bpDataPage = processDataForWhereClause( dfwcp );
			
				if( bpDataPage == NULL ){
						
					continue;
						
				}
			

				// (c) pair "bpDataPage" with corresponding "pageID" ..
				if ( bpDataPage->getNumberOfRows() > 0 ) {
					
					if ( getBufferManager()->getBufferInfoTable()->getPageID()->at(i) < 0 ) {
					
						// since BP pages that have yet to be written to file do not yet
						// have a pageID, where "bm->bit.pageID[i] == -1",
						// and since the "map" container "bpReturnPages" cannot duplicate keys == -1,
						// instead use "bpOffset" as the key for the "bpDataPage".
						// i.e.: i = 0, bpOffset = -1
						//		 i = 1, bpOffset = -2, etc ..
					
						int bpOffset = ( (-1) * (i+1) );
						
						bpReturnPages->insert( pair< int, DataPage* >( bpOffset, bpDataPage ) );

					
					} else {
					
						bpReturnPages->insert( pair< int, DataPage* >( getBufferManager()->getBufferInfoTable()->getPageID()->at(i), bpDataPage ) );
						

					} // end "if()" ..
					
					
				} // end "if ( bpDataPage.records.size() > 0 )" ..
				

				bitIndicesForNonEmptyBPBlocks->push_back( i );
				
				bpBlocksFound = true;
				

					
			} // end "if ( ( !bm->bit.empty[ i ] ) && " ..
			
			
		} // end "for()" ..
		
		//cout << "QueryManager:: selectFromX() Point K.." << endl << endl;
	
		
		// 7) check for pages from the file for this "tableName" ..
		if ( fileHasPages ) {
			
			//cout << "QueryManager:: selectFromX() Point K1.." << endl << endl;
			
			vector<int>* returnPageIDs = new vector<int>();
				
			bool filePageFoundInBP = false;
			
			int bitIndex = -1;
			
			// first, get out the "attribute names" from the "where" clause which have indices ..
			// (consolidate the "attribute names" from "leftName", "rightName" and "columnName
			
			vector<string>* whereClauseColumnNames = new vector<string>();
			vector<int>* whereClauseColumnValues = new vector<int>();
			string attributeName = "";
			int attributeValue = -1;
			bool hasIndex = false;
			int columnNameSize = columnName->size();
				
			for(int i=0; i<columnNameSize; i++){
					
				attributeName = columnName->at(i);
				
				hasIndex = getCatalogManager()->doesCatalogTableAttributeNameHaveIndex( tableNameX, attributeName );
					
				if( hasIndex == true ){
						
					whereClauseColumnNames->push_back( attributeName );
					
					attributeValue = columnValue->at(i);
					
					whereClauseColumnValues->push_back( attributeValue );
					
				}
						
			}
			
			//cout << "QueryManager:: selectFromX() Point K2.." << endl << endl;
			
			// pageIds used to collect "fileDataPages" ..
			vector<int>* masterPageIds = new vector<int>();
			
			// one or more attributes have an index ..
			if( hasIndex == true ){
				
				//cout << "QueryManager:: selectFromX() Point K3.." << endl << endl;
				
				// need pageIds from Index if one (or more) of the "where" clause attributes has an index ..
				vector< vector<int>* >* pageIds = new vector< vector<int>* >();
				vector<int>* indexPageIds = new vector<int>();
				int columnNameIndex = -1;
				string indexName = "";
				int whereClauseColumnNamesSize = whereClauseColumnNames->size();
				for(int i=0; i<whereClauseColumnNamesSize; i++){
				
					attributeName = whereClauseColumnNames->at( i );
					
					//cout << "QueryManager:: selectFromX() Point K4.." << endl << endl;
				
					indexName = getCatalogManager()->getCatalogIndexName( tableNameX, attributeName );
					
					//cout << "QueryManager:: selectFromX() Point K5.." << endl << endl;
				
					attributeValue = whereClauseColumnValues->at( i );
					
					//cout << "QueryManager:: selectFromX() attributeName: " << attributeName << ", indexName: " << indexName << ", attributeValue: " << attributeValue << endl << endl;
					
					indexPageIds = getIndexManager()->lookUpIDs( tableNameX, indexName, attributeValue );
				
					pageIds->push_back( indexPageIds );
				
				}
				
				//cout << "QueryManager:: selectFromX() Point K6.." << endl << endl;
			
				// go through "pageIds" and remove duplicate values, putting the result in "masterPageIds" ..
				int pageIdValue = -1;
				vector<int>* currentPageIds = new vector<int>();
				int pageIdsSize = pageIds->size();
				int currentPageIdsSize = -1;
				for(int i=0; i<pageIdsSize; i++){
				
					currentPageIds = pageIds->at(i);
				
					currentPageIdsSize = currentPageIds->size();
				
					for(int j=0; j<currentPageIdsSize; j++){
					
						pageIdValue = currentPageIds->at(j);
						
						//cout << "QueryManager:: selectFromX() pageIdValue: " << pageIdValue << endl << endl;
					
						if( contains( pageIdValue, masterPageIds ) == false ){
						
							masterPageIds->push_back( pageIdValue );
						
						}
					
					}

				}
				
				//cout << "QueryManager:: selectFromX() Point K7.." << endl << endl;
			
			// there is no index ..
			} else {
				
				//cout << "QueryManager:: selectFromX() Point K8.." << endl << endl;
				
				// so just use the pageIds from the "DataHeader" for this tableNameX ..
				
				for(int i=0; i < header->getTotalNumberOfDataPages(); i++){
					
					masterPageIds->push_back( header->getHeaderEntries()->at( i )->getPageID() );
					
				}
				
			} //  end "if()" ..
			
			
			
			//cout << "QueryManager:: selectFromX() Point L.." << endl << endl;
			
			// (a) check the DataHeader and collect all the "pageID"s for this "tableName",
			//     which are _not_ already in the BP ..
			int masterPageIdsSize = masterPageIds->size();
			
			//cout << "QueryManager:: selectFromX() masterPageIdsSize: " << masterPageIdsSize << endl << endl;
			
			
			for(int i=0; i<masterPageIdsSize; i++){
		
				// (i) if bp blocks were found above, this indicates that file pages matching
				//     the bp blocks found should be skipped ..
				if ( bpBlocksFound ) {
					
					for (int j=0; j < bitIndicesForNonEmptyBPBlocks->size(); j++){

						bitIndex = j;
					
						if( masterPageIds->at( i ) == getBufferManager()->getBufferInfoTable()->getPageID()->at( bitIndicesForNonEmptyBPBlocks->at(j) ) ){
					
							// pageID from "HP" matches pageID in the "bit" for "non-empty" BP blocks, 
							// so try another pageID from the "HP" ..
							filePageFoundInBP = true;
							break;
					
						}
					
					} // end "for()" ..
					
					if ( !filePageFoundInBP ) {
						
						// "pageID" from the "HP" was not found in the BP,
						// so go ahead and collect it as a valid entry to return
						// to the user ..
						
						returnPageIDs->push_back( masterPageIds->at( i ) );
						
						
					} // end "if ( !filePageFoundInBP )" ..
					
					// reset this parameter for the next iteration ..
					filePageFoundInBP = false;
					
					
				// (ii) There were no bp blocks found above, so 
				//      only collect DataHeader pageID ..
				} else {
					
					// from the file, collect all the "pageID"s for this "tableName" ..
					returnPageIDs->push_back( header->getHeaderEntries()->at( i )->getPageID() );
					
				}
				
		
			} // end "for()" ..
			
/*
			for(int i=0; i < header->getTotalNumberOfDataPages(); i++){
		
				// (i) if bp blocks were found above, this indicates that file pages matching
				//     the bp blocks found should be skipped ..
				if ( bpBlocksFound ) {
					
					for (int j=0; j < bitIndicesForNonEmptyBPBlocks->size(); j++){

						bitIndex = j;
					
						if( header->getHeaderEntries()->at( i )->getPageID() == getBufferManager()->getBufferInfoTable()->getPageID()->at( bitIndicesForNonEmptyBPBlocks->at(j) ) ){
					
							// pageID from "HP" matches pageID in the "bit" for "non-empty" BP blocks, 
							// so try another pageID from the "HP" ..
							filePageFoundInBP = true;
							break;
					
						}
					
					} // end "for()" ..
					
					if ( !filePageFoundInBP ) {
						
						// "pageID" from the "HP" was not found in the BP,
						// so go ahead and collect it as a valid entry to return
						// to the user ..
						
						returnPageIDs->push_back( header->getHeaderEntries()->at( i )->getPageID() );
						
						
					} // end "if ( !filePageFoundInBP )" ..
					
					// reset this parameter for the next iteration ..
					filePageFoundInBP = false;
					
					
				// (ii) There were no bp blocks found above, so 
				//      only collect DataHeader pageID ..
				} else {
					
					// from the file, collect all the "pageID"s for this "tableName" ..
					returnPageIDs->push_back( header->getHeaderEntries()->at( i )->getPageID() );
					
				}
				
		
			} // end "for()" ..
			
*/
			
			//cout << "QueryManager:: selectFromX() Point M.." << endl << endl;
			
			//cout << "QueryManager:: selectFromX() tableNameX: " << tableNameX << endl << endl;
			
			// (b) loop through the all the "pageID"s for this "tableName",
			//     collect the associated "dataPage"s in order to return them ..
			int returnPageIDsSize = returnPageIDs->size();
			for(int i=0; i < returnPageIDsSize; i++){
			
				fileDataPage = getBufferManager()->returnFilePage( tableNameX, returnPageIDs->at(i) );
				
				//cout << "QueryManager:: selectFromX() Point M1.." << endl << endl;
				//cout << "QueryManager:: selectFromX() fileDataPage-getNumberOfEntries: " << fileDataPage->getNumberOfEntries() << endl << endl;
				//fileDataPage->printDataPage();
				
				if ( fileDataPage == NULL ) {
				
					continue;
				
				}
				
				// if "returnFilePage()" fails to return any file pages, 
				// continue checking the remaining pages ..
				if ( fileDataPage->getNumberOfRows() == 0 ) {
				
					continue;
				
				}
				
				// need to set all parameters (not just "dataPage"),
				// because may have only "fileDataPage" or just "bpDataPage",
				// (if you had both, these would have been set above for "bpDataPage") ..
				dfwcp->setDataPage( fileDataPage );
				dfwcp->setQueryPairs( queryPairs );
				dfwcp->setIsPairs( isPairs );
				dfwcp->setTableNameX( tableNameX );
				dfwcp->setDoesQueryHaveWhereClause( doesQueryHaveWhereClause );
				dfwcp->setQuerySpecifiesColumnNames( querySpecifiesColumnNames );
				dfwcp->setColumnNames( columnNames );
				dfwcp->setMaxNumberOfAttributes( maxNumberOfAttributes );

				// (b) process data according to "where" clause criteria ..
				fileDataPage = processDataForWhereClause( dfwcp );
				
				//cout << "QueryManager:: selectFromX() fileDataPage-getNumberOfEntries: " << fileDataPage->getNumberOfEntries() << endl << endl;
				
				if( fileDataPage == NULL ){
					
					continue;	
					
				}
				
				if ( fileDataPage->getNumberOfRows() > 0 ) {
					
					// only insert into "fileReturnPages"
					// if the "fileDataPage" has any records ..
					fileReturnPages->insert( pair< int, DataPage* >( returnPageIDs->at(i), fileDataPage ) );
					
				}

			} // end "for()" ..

		} // end "if ( fileHasPages )" ..
		
		//cout << "QueryManager:: selectFromX() Point N.." << endl << endl;
		
		// cleanup ..
		delete dfwcp;
		
		//cout << "QueryManager:: selectFromX() Point N1.." << endl << endl;
		
/*
		// DEBUGGING ..
		if( bpReturnPages != NULL){
			
			//vector< vector<int>* >* entries = NULL;
			map<int, DataPage*>::iterator it;
			for(it=bpReturnPages->begin(); it!=bpReturnPages->end(); it++){
				
				bpDataPage = it->second;
				//bpDataPage = bpReturnPages[i];
				
				if( bpDataPage != NULL )
					bpDataPage->printDataPage();
				
				// end of page ..
				cout << "--------" << endl;
				
			} // end "for()" ..
			
		}
		
		// DEBUGGING ..
		if( fileReturnPages != NULL){
			
			//vector< vector<int>* >* entries = NULL;
			
			map<int, DataPage*>::iterator it;
			for(it=fileReturnPages->begin(); it!=fileReturnPages->end(); it++){
				
				fileDataPage = it->second;
				//fileDataPage = fileReturnPages[i];
				
				if( fileDataPage != NULL )
					fileDataPage->printDataPage();
				
				// end of page ..
				cout << "--------" << endl;
				
			} // end "for()" ..
			
		}
*/
		
		//cout << "QueryManager:: selectFromX() Point N2.." << endl << endl;
		
		vector< map<int, DataPage* >* >* returnPages = new vector< map<int, DataPage* >* >();
		
		//cout << "QueryManager:: selectFromX() Point N3.." << endl << endl;
		
		returnPages->push_back( bpReturnPages );
		
		//cout << "QueryManager:: selectFromX() Point N4.." << endl << endl;
		//cout << "QueryManager:: selectFromX() bpReturnPages-size: " << bpReturnPages->size() << endl << endl;
		
		returnPages->push_back( fileReturnPages );
		
		//cout << "QueryManager:: selectFromX() Point N5.." << endl << endl;
		//cout << "QueryManager:: selectFromX() fileReturnPages-size: " << fileReturnPages->size() << endl << endl;
		
		return returnPages;

		
	} else {
	 
	 	// the "tableNameX" does not exist in the Catalog,
	 	// so return an error message ..
	 	
	 	errorString = "\nThe table ";
	 	errorString += tableNameX;
	 	errorString += " does not exist [ QM::selectFromX ].";
	 	
	 	// update the "MessageManager" to track this error and output to user ..
		getMessageManager()->setErrorStringQM( errorString );	// string detailing the error in "QueryManager" ..
		
		vector< map<int, DataPage* >* >* returnPages = new vector< map<int, DataPage* >* >();
		
		return returnPages;
		
	}
	
	//cout << "QueryManager:: selectFromX() Point P.." << endl << endl;
	
	
} // end "selectFromX()" ..


map<int, string>* QueryManager:: groupSelectColumnNames( string tableNameX, vector<string>* selectColumnNames ){
	
	// match valid column names with their position order (from left to right)
	// as it is recorded in the catalog ..
	
	//cout << "QueryManager:: groupSelectColumnNames() Point A.." << endl << endl;
	
	if( selectColumnNames == NULL )
		return NULL;
		
	if( tableNameX.length() <= 0 )
		return NULL;
	
	// DEBUGGING ..
	//for(int i=0; i<selectColumnNames->size(); i++){
	//	cout << "QueryManager:: groupSelectColumnNames() selectColumnNames: " << selectColumnNames->at(i) << endl << endl;	
	//}
	
	// DEBUGGING ..
	//int columnNamesSize = -1;
	//if( columnNames != NULL )
	//	columnNamesSize = columnNames->size();
	//for(int i=0; i<columnNamesSize; i++)
	//	cout << "QueryManager:: selectFromX() columnNames->at(i("<<i<<")): " << (*columnNames)[i] << endl << endl;		

		
	//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() start .." << endl << endl;
		
	int columnNamePosition = -1;
		
	map<int, string>* columnNames = new map<int, string>();
		
	vector<int>* processedColumnNamePositions = new vector<int>();
	
	bool tryNextSelectColumnName = false;
	
	string selectColumnName = "";
	
	int processedColumnNamePositionsSize = -1;
	
	//if( selectColumnNames == NULL )
	//	return NULL;
		
	bool doesAttributeExist = false;
	
	//cout << "QueryManager:: groupSelectColumnNames() Point B.." << endl << endl;
	
	// linear scan of "selectColumnNames" in order to match
	// "columnNamePosition" with "selectColumnNames[i]" ..
	int selectColumnNamesSize = selectColumnNames->size();
	
	for(int i=0; i<selectColumnNamesSize; i++){
			
		//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point A[i("<<i<<")].." << endl << endl;
		
		
		//cout << "QueryManager:: groupSelectColumnNames() Point C[i("<<i<<")].." << endl << endl;
		
		selectColumnName = selectColumnNames->at(i);
			
		doesAttributeExist = getCatalogManager()->doesCatalogTableAttributeNameExist( tableNameX, selectColumnName );
			
		//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() doesAttributeExist("<<selectColumnName<<"): " << doesAttributeExist << endl << endl;
			
		if ( doesAttributeExist ) {
				
			//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point B[i("<<i<<")].." << endl << endl;
				
			// 1) if the attributeName is valid for this tableName, continue ..
			
			//cout << "QueryManager:: groupSelectColumnNames() Point D[i("<<i<<")].." << endl << endl;
				
			selectColumnName = selectColumnNames->at(i);

			if( selectColumnName.length() <= 0 )
				continue;
				
			//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point C[i("<<i<<")].." << endl << endl;
				
			columnNamePosition = getCatalogManager()->getAttributePosition( tableNameX, selectColumnName );
			
			//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() columnNamePosition("<<selectColumnName<<"): " << columnNamePosition << endl << endl;
			
			if ( columnNamePosition >= 0 ) {
				
				//cout << "QueryManager:: groupSelectColumnNames() Point E[i("<<i<<")].." << endl << endl;
						
				// (a) if columnNamePosition != -1 (columnName not found), continue ..
					
				//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point D[i("<<i<<")].." << endl << endl;
					
				processedColumnNamePositionsSize = processedColumnNamePositions->size();
					
				for(int j=0; j<processedColumnNamePositionsSize; j++){
						
					if ( columnNamePosition == processedColumnNamePositions->at(j) ){
							
						// this "columnNamePosition" was seen / processed already,
						// so skip it and try the next one ..
						tryNextSelectColumnName = true;
						break;
						
					}
						
				}
				
				//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() tryNextSelectColumnName("<<selectColumnName<<"): " << tryNextSelectColumnName << endl << endl;
				
				if ( tryNextSelectColumnName ) {
					
					// (b) since this "columnNamePosition" was found in
					//     "processedColumnNamePositions[j]", 
					//     skip it and try the next one ..
					continue;
					
				}
					
				// (c) if the columnName hasn't already been seen / processed, 
				//     match columnName with its columnNamePosition ..
					
				//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point E[i("<<i<<")].." << endl << endl;
					
				//cout << "QueryManager:: groupSelectColumnNames() columnNamePosition: " << columnNamePosition << ", selectColumnName: " << selectColumnName << endl << endl;
					
				columnNames->insert( pair<int, string>( columnNamePosition, selectColumnName ) );
				
				//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point F[i("<<i<<")].." << endl << endl;
				
				processedColumnNamePositions->push_back( columnNamePosition );
					
				// (d) reset this parameter for the next iteration ..
				columnNamePosition = -1;
				
				//cout << "QueryManager:: groupSelectColumnNames() Point F[i("<<i<<")].." << endl << endl;
				
				//cout << "QueryManager:: selectFromX() >> groupSelectColumnNames() Point G[i("<<i<<")].." << endl << endl;
				
				
			} // end "if ( columnNamePosition >= 0 )" ..
				
				
		} // end "if ( bm->cm->doesCatalogTableAttributeNameExist() )" ..

			
	} // end "for()" ..
	
	//cout << "QueryManager:: groupSelectColumnNames() Point G.." << endl << endl;
	
	//cout << "QueryManager:: groupSelectColumnNames() columnNames->size(): " << columnNames->size() << endl << endl;
		
	return columnNames;

	
} // end "groupSelectColumnNames( string tableName, vector<string> selectColumnNames )" ..
