#ifndef USERINTERFACE_H_
#define USERINTERFACE_H_

#pragma once

#include "../Catalog/CatalogFile.h"

#include "../Database/DatabaseFile.h"

#include "../Utilities/String_Utilities.h"
#include "../Utilities/TableFormatter.h"
#include "../Utilities/Timer.h"

#include "../Managers/BufferManager.h"
#include "../Managers/QueryManager.h"
#include "../Managers/StorageManager.h"
#include "../Managers/CatalogManager.h"
#include "../Managers/MessageManager.h"

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <cstring>


using namespace std;

class UserInterface {
	
public:

	UserInterface( int maxColumnsPerTable, int bufferPoolSize );
	~UserInterface();
	
	QueryManager* getQueryManager();
	
	IndexManager* getIndexManager();
	
	BufferManager* getBufferManager();
	
	StorageManager* getStorageManager();
	
	CatalogManager* getCatalogManager();
	
	MessageManager* getMessageManager();
	
	void printPath();							// print current fileDrive + fileDirectory path to console ..
	
	void runBatchCommands();					// run a batch file of DB commands ..
	
	void promptUserForCommand();				// prompt the user for a command ..

	void preProcessCommand();					// check for "missing" command or "logout" -type keyword

	void parseCommand();						// determine if the command matches a 
												// known command ..
	
	void performCommand();						// perform known commands ..
	
	void printDatabaseTable();					// helper method, used in "performCommand()" ..
	
	void printMessagesAndCleanUp();				// print error or result messages (if any)
												// and reset parameters ..
	
	void printResults( map<int, DataPage* >* bpDataPages,  map<int, DataPage* >* fileDataPages, vector<string>* selectColumnNames );
	
	void printBufferResults();					// print the results of the command "print buffer" ..
	
	void printBufferDataResults();				// print the results of the command "print buffer data" ..
	
	bool isQuitProgram();						// get status, whether to quit program or not ..


	// utility method ..
	void addTableName( string tableName_ );		// used for testing ..
	void clearTableNames();						// used for testing ..
	

	// Following methods not currently used ..
	void clearOldTablePageLocks();
	void oldPrintBufferResults();
	void oldPrintResults( map<int, DataPage* >* bpDataPages,  map<int, DataPage* >* fileDataPages, vector<string>* selectColumnNames );

	void changePath( vector<string>* inputTokens_, bool isWindows );		// tester method for setting the path prior to running test ..

/*
 * SETTERS ..
 */

	void resetValues();										// reset tracking the identified first fields from commandInput ..

	void setInputTokens( vector<string>* inputTokens_ );	// this is a "tester" method ..
	
	void setTableName( string tableName_ );
	
	void setSelectColumnNames( vector<string>* selectColumnNames_ );
	
	void setWhereColumnNamePairs( multimap<string, string>* whereColumnNamePairs_ );
	
	void setWhereColumnNameValuePairs( map<string, int>* whereColumnNameValuePairs_ );

	
/*
 * GETTERS ..
 */
 
	string getTableName();									// for use in "validateInsertValues()" ..
	
	vector<string>* getSelectColumnNames();
	
	multimap<string, string>* getWhereColumnNamePairs();
	
	map<string, int>* getWhereColumnNameValuePairs();


private:

	bool quitProgram;										// flag indicating whether or not to end this program ..

	string batchFileName;									// string representing the file name of the batch file ..

	string commandInput;									// string representing user command input ..

	vector<string>* inputTokens;							// vector containing tokens of commandInput ..
	

	string errorString;										// string to set and output to user when error occurs ..

	bool hasParserError;									// flag indicating whether or not the parser has generated an error ..


	string firstField;										// track the first fields from the commandInput that was identified ..

	string secondField;										// track the second fields from the commandInput that was identified ..

	string thirdField;										// track the third fields from the commandInput that was identified ..
	
	string fourthField;										// track the fourth fields from the commandInput that was identified ..
	
	string fifthField;										// track the fifth fields from the commandInput that was identified ..

	string sixthField;										// track the sixth fields from the commandInput that was identified ..

	string seventhField;									// track the seventh fields from the commandInput that was identified ..
	
	string eighthField;										// track the eighth fields from the commandInput that was identified ..

	
	string tableName;										// indicates the tableName (for "create" and "insert" statements only) ..
	
	vector<string>* tableNames;								// vector containing tableName (for "create", "insert" or "select" statements) ..

	int numberOfTables;										// track the number of tables (for "select" statement only) ..

	string indexName;										// indicates the indexName (for "create" and "insert" statements only) ..


	vector<string>* createColumnNames;						// vector containing columnNames (for "create" statement only) ..

	bool isNoDuplicates;									// used for "parseCreateStatement" for Index, where "NO DUPLICATES" keyword sets isNoDuplicates == true ..

	int numberOfColumns;									// track the number of columns (for "create", "insert" and "select" statements only) ..
	
	
	vector<string>* insertValues;							// vector containing inserValues (for "insert" statement only) ..
	
	vector<string>* selectColumnNames;						// vector containing columnNames (for "select" statement only) ..
	

	multimap<string, string>* whereColumnNamePairs;			// multimap containing both columnNames (LHS + RHS of expression) (for statements with "where" clauses only) ..
	
	map<string, int>* whereColumnNameValuePairs;			// map containing both columnValues (LHS + RHS of expression) (for statements with "where" clauses only) ..
	
	
	Timer timer;											// object used for timing the DB operations, via the "set timer on" command ..
	
	string fileDrive;										// default root drive where program expects to read/write ..
	
	string fileDirectory;									// default directory where program expects to read/write ..
	
	int fileDirectoryStartPosition;							// used for "parseSetStatement()" and "performCommand() -> set path" ..
	
	bool isWindows;											// used for "parseSetStatement()" and "performCommand() -> set path" ..
	
	
	// CONSTANTS ..
	const int MAX_COLUMNS_PER_TABLE;						// number of columns per table max limit ..
	
	
	QueryManager* qm;										// This is the class that the parser uses to interact with the buffer, catalog and database ..
	
	
	// non-static member functions ..
	bool parseFirstField();									// parse and identify first fields in command line input ..
	
	bool hasSecondField();									// determine if the commandInput has a second fields ..

	bool parseRemainingFields();							// parse remaining fields in command line input ..
	
	void parseCreateStatement();							// perform parse on "create" statement ..
	
	void parseInsertStatement();							// perform parse on "insert" statement ..
	
	void parseSelectStatement();							// perform parse on "select" statement ..
	
	void parsePrintStatement();								// perform parse on "print" statement ..
	
	void updateFileDirectoryPaths();						// helper method for "performCommand()" ..
	
	void updatePathOperatingSystem();						// helper method for "parseSetStatement()" ..
	
	void parseSetStatement();								// perform parse on "set" statement ..
	
	void parseShowStatement();								// perform parse on "show" statement ..
	
	void parseRunStatement();								// perform parse on "run" statement ..
	
	bool validateKeywordField( int, vector<string> );		// identify keyword fields (for non-"select" statements only) in command line input ..
	
	bool validateKeywordFieldSelect( int, vector<string> ); // identify keyword fields (for "select" statement only) in command line input ..
	
	bool validateField( int fieldNumber );					// identify non-keyword fields (i.e.: columnName(s), tableName(s), etc) ..
	
	bool validateTableName( int fieldNumber );				// identify tableName (for "create" and "insert" statements only) ..
	
	bool validateTableNames( int fieldNumber );				// identify tableName (for "select" statement only) ..
	
	bool validateIndexName( int fieldNumber );				// identify indexName (for "create" and "insert" statements only) ..
	
	bool validateColumnNamesCreate( int fieldNumber ); 		// identify columnNames (for "create" (table) statement only) ..

	bool validateColumnNamesCreateIndex( int fieldNumber ); // identify columnNames (for "createIndex" statement only) ..

	bool validateInsertValues( int fieldNumber ); 			// identify insert values (for "insert" statement only) ..
	
	bool validateColumnNamesSelect( int fieldNumber );		// identify select values (for "select" statement only) ..
	
	bool validateColumnNamesWhere( int fieldNumber );		// identify select values (for statements with "where" clauses only) ..
	
	bool validatePathName( int fieldNumber );				// identify path name (fileDrive + fileDirectory) (for "set" statement only) ..
	
	bool validateBatchFilePathName( int fieldNumber );		// identify batch file path name (for "run" statement only) ..
	
}; // end of "UserInterface" class ..


#endif /*USERINTERFACE_H_*/
