#ifndef ARRAY_H_
#define ARRAY_H_

#pragma once

#include <iostream>
#include <cassert>

using namespace std;

template<class T>
class Array {
	
	// NOTE: for some reason, compiler doesn't recognize the overloaded definitions
	// for these operators made in this class (these operators don't work on this object) .. ?
	friend ostream& operator<< (ostream&, const Array<T>&);
	friend istream& operator>> (istream&, Array<T>&);

public:

	// constructors ..
	Array();								// [d] default ..
	Array(int);								// [d] constructor with size ..
	Array(int, T[]);						// [d] constructor with size, initialization ..
	Array(const Array&);					// [d] copy constructor ..
	~Array();								// [d] destructor ..
	
	// getters ..
	int getSize() const;					// [d] return Array size
	static int getArrayCount();				// [d] return count of Array's instantiated
											//     (tracks the number of Array's objects)
											
	// setters ..
	void setSize(int);						// [d] initialize and/or change the size of the Array ..
	void setValues(T[]);					// [d] initialize and/or change the values in the Array ..
	void setValues(int, T[]);				// [d] initialize and/or change the values in the Array ..	

	// manipulate element functions ..
	void insertElement(T, int);				// [d] insert an element into Array, at position passed in (increase Array size by 1) ..
	void replaceElement(T, int);			// [d] replace element in Array, at position passed in (do not resize Array) ..
	T& removeElement(int);					// [d] remove an element from the array, at position passed in (decrease Array size by 1)..
	T& getElement(int);						// [d] retrieve an element from the array, from position passed in (leave element unchanged) ..
	void pushElement(T);					// [d] add an element at the end of the array (increase Array size by 1) ..
	T& popElement();						// [d] remove the element at the end of the array (decrease Array size by 1)..
	
	// overloaded operators ..
	const Array& operator= (const Array<T> &); 		// [d] assign arrays
	bool operator== (const Array<T> &) const;		// [d] compare equals
	bool operator!= (const Array<T> &) const;		// [d] compare not equal
	T& operator[] (int);							// [d] subscript operator
	
	// utility functions ..
	void print() const;						// [d] print contents of Array to console ..
	void print(char*) const;				// [d] print contents of Array to console, input prompt to use (i.e.: print("contents") yields contents[0] = x, contents[1] = x, .. etc) ..
	void sort();							// [d] re-order the elements in sorted ("quicksort") order ..
	void reverse_sort();					// [d] re-order the elements in reverse-sorted ( reverse "quicksort") order ..
	void merge(const Array&);				// [d] merge another Array with this Array, sort ("quicksort") the result ..
	void flip();							// [d] flip the order of the elements ..
	int search(T);							// [d] search ("Binary Search") and find the position of an element in the Array ..

private:

	// non-static data members ..
	T* ptr;									// pointer to 1st element of Array ..
	int size;								// size of the Array ..
	
	// static data members ..
	static int arrayCount;					// number of Array's instantiated ..
	
	// non-static member functions ..
	int split(int, int);					// find the "half-way" point in the array ..
	void quicksort(int, int);				// recursive quicksort method ..
	
	int reverse_split(int, int);			// find the "half-way" point in the array ..
	void reverse_quicksort(int, int);		// recursive quicksort method ..

}; // end of "Array" class ..


/*
 * "STATIC DATA MEMBERS" ..
 *  initialize "static data member" at file scope ..
 */

template<class T>
int Array<T>:: arrayCount = 0;				// start with no instantiated objects ..


/*
 * CONSTRUCTORS ..
 */
 
 template<class T>
Array<T>:: Array(){
	
	size = 0;								// size is not yet set ..
	
	++arrayCount;							// increment static object count ..
	
} // end of "Array()" ..


template<class T>
Array<T>:: Array(int arraySize){
	
	ptr = new T[arraySize];					// create space for new Array, point to it ..
	
	assert(ptr != 0);						// terminate if memory not allocated ..
	
	size = arraySize;						// set size for this object ..
	
	++arrayCount;							// increment static object count ..
	
	/*
	 * NOTE: unable to initialize this object.
	 * "class T" is unknown until compile-time.
	 * The argument(s) to the constructor of "class T"
	 * are also unknown, which makes it impossible
	 * to properly instantiated objects of "class T"
	 * that would then be used to initialize this object.
	 */
	
} // end of "Array(int arraySize)" ..



template<class T>
Array<T>:: Array( int arraySize, T initialValues[] ){
	
	ptr = new T[arraySize];							// create space for new Array, point to it ..
	
	assert(ptr != 0);								// terminate if memory not allocated ..
	
	size = arraySize;								// set size for this object ..
	
	++arrayCount;									// increment static object count ..
	
	// if (size of initialValues[]) > arraySize, 
	// some of the initialValues[] wont be used, and
	// if (size of initialValues[]) < arraySize,
	// some of the Array[arraySize] wont be initialized ..
	
	// correctness in initializing is left up to the user of this class ..
	
	for(int i=0; i<size; i++){						// initialize new Array ..
		
		ptr[i] = initialValues[i];					// set the values ..
			
	}
	
} // end of "Array(int arraySize, T[] initialValues)" ..


template<class T>
Array<T>:: Array(const Array<T>& initArray){
	
	ptr = new T[initArray.size];					// create space for new Array, point to it ..
	
	assert(ptr != 0);								// terminate if memory not allocated ..
	
	size = initArray.size;							// set size for this object ..
	
	++arrayCount;									// increment static object count ..
	
	for(int i=0; i<size; i++){						// initialize new Array ..
		
		ptr[i] = initArray.ptr[i];					// set the values ..
			
	}
	
	
} // end "Array(const Array& initArray)" ..


/*
 * DESTRUCTOR ..
 */

template<class T>
Array<T>:: ~Array(){
	
	delete [] ptr;									// deallocate space for Array ..
	
	--arrayCount;									// decrement static object count ..

	
} // end of "~Array()" ..


/*
 * GETTERS ..
 */

template<class T>
int Array<T>:: getSize() const {
	
	return size;	
	
}


template<class T>
int Array<T>:: getArrayCount(){
	
	return arrayCount;
	
} // end "getArrayCount()" ..


/*
 * SETTERS ..
 */

template<class T>
void Array<T>:: setSize( int newSize ) {
	
	bool isResize = true;
	
	int swapSize = 0;
	
	// size == newSize, 
	// so don't change Array ..
	if (size == newSize) {
	
		// don't bother to swap arrays ..
		isResize = false;
	
	} else if ( size == 0 ) {
		
		// this indicates that the "empty constuctor" was used,
		// so go ahead with the remainder of "constructor" operations ..
		
		// don't bother to swap arrays ..
		isResize = false;
		
		ptr = new T[newSize];						// create space for new Array, point to it ..
	
		assert(ptr != 0);							// terminate if memory not allocated ..
	
		size = newSize;								// set size for this object ..
		
	} else if( size < newSize ){
		
		// size < newSize, then the size of elements of Array to swap
		// is the current Array[size] into the new Array[newSize] ..
		swapSize = size;
		 
	// size > newSize, then the size of elements of Array to swap
	// is the new Array[newSize] is the max number of elements to keep from Array[size] ..
	} else if( ( size > newSize ) || ( size == 0 ) ){
		
		swapSize = newSize;
		
	// size == 0, then the Array hasn't had all Constructor operations conducted yet ..
	}
	
	// only resize the Array if newSize != size ..
	if ( isResize ) {
		
		T* newPtr = new T[newSize];					// create space for new Array, point to it ..
	
		assert(newPtr != 0);						// terminate if memory not allocated ..
	
		assert(ptr != 0);							// terminate if memory not allocated ..
	
		for(int i=0; i<swapSize; i++){				// swap data from Array[size] into Array[newSize] ..
			
			newPtr[i] = ptr[i];
			
		}
		
		delete [] ptr;								// deallocate space for Array ..
		
		ptr = newPtr;								// pointer points at new Array ..
	
		size = newSize;								// set new size for this object ..
		
	}
	
} // end of "setSize(int newSize)" ..


template<class T>
void Array<T>:: setValues( T newValues[] ) {
	
	assert(ptr != 0);
	
	//assert(initialValues != 0);
	
	// only initialize with values if 
	// the size is already set ..
	if( size > 0 ) {
		
		// if (size of newValues) > size, 
		// Array[size] is filled up and remainder of newValues is left off (unused) ..
		// else if (size of newValues) < size,
		// Array[size] is partially filled up to (size of newValues), 
		// and remainder of Array is over-written with garbage values (dropping off the end of "newValues") ..
		
		// correctness in setting Array values is left up to the user of this class ..
		
		for(int i=0; i<size; i++){						// swap data from initialValues[size] into Array[size] ..
			
			ptr[i] = newValues[i];
			
		}
		
	} 
	
} // end of "setValues( T initialValues[] )" ..


template<class T>
void Array<T>:: setValues( int newSize, T* newValues ) {
	
	assert(ptr != 0);
	
	// only initialize with values if 
	// the size is already set ..
	if( size > 0 ) {
		
		// correctness in setting Array values is left up to the user of this class ..
		
		// if newSize > size, 
		// Array[size] is filled up and remainder of newValues is left off ..
		if( newSize > size ) {
			
			delete [] ptr;									// deallocate space for Array ..
			
			ptr = newValues;								// swap Array[size] for Array[newSize] ..
			
			assert(ptr != 0);
			
			size = newSize;									// set new size for this object ..
			
		// else if newSize < size,
		// Array[size] is partially filled up to newSize, and remainder of Array is unchanged ..
		// also, if newSize == size,
		// Array[size] is replaced with the contents of newValues ..
		} else if ( newSize <= size ) {
			
			for(int i=0; i<newSize; i++){					// swap data from newValues[size] into Array[size] ..
			
				ptr[i] = newValues[i];
			
			}
			
		}
		

		
	} 
	
} // end of "setValues( int newSize, T initialValues[] )" ..


/*
 * MANIPULATE ELEMENT FUNCTIONS ..
 */
 

template<class T>
void Array<T>:: insertElement( T element, int insertPosition ){
	
	// DEBUGGING ..
	//cout << "insertElement - element: " << element << endl;
	
	if( size == 0 ) {
		
		// DEBUGGING ..
		//cout << "insertElement (size == 0) - insertPosition: " << insertPosition << endl;
		
		ptr = new T[1];									// create space for new Array, point to it ..
	
		assert(ptr != 0);								// terminate if memory not allocated ..
	
		size = 1;										// set size for this object ..
	
		ptr[0] = element;								// insert the value ..
		
	} else if( (0 <= insertPosition) && (insertPosition < size) ) {
		
		// DEBUGGING ..
		//cout << "insertElement (insertPosition(" << insertPosition << ") < size(" << size << ")) - insertPosition: " << insertPosition << endl;
		
		T* newPtr = new T[size+1];
		
		assert(newPtr != 0);
		
		// account for translating between previous array
		// to new array (following insert operation),
		// where new array has one additional cell ..
		int insertSpace = 0;
		
		++size;
		
		for(int i=0; i<size; i++){
			
			if ( i == insertPosition ) {
				
				newPtr[i] = element;
				
				insertSpace++;
				
				//cout << "insertElement (insertPosition < size): newPtr[" << i << "](" << newPtr[i] << ") = element(" << element << ")" << endl;
				
			} else {
				
				// before insert operation: (i - insertSpace) = i - 0
				//  after insert operation: (i - insertSpace) = i - 1 ..
				
				newPtr[i] = ptr[(i - insertSpace)];
				
				//cout << "insertElement (insertPosition < size): newPtr[" << i << "](" << newPtr[i] << ") = ptr[" << (i - insertSpace) << "](" << ptr[(i - insertSpace)] << ")" << endl;
				
			}
			
		} // end "for()" ..
		
		
		// update "ptr" with new values ..
		delete [] ptr;
		
		ptr = newPtr;
		
		
	} // end "if( size == 0 )" ..
	
	
} // end "insertElement( T object, int insertPosition )" ..


template<class T>
void Array<T>:: replaceElement( T element, int replacePosition ){
	
	// DEBUGGING ..
	//cout << "replaceElement - element: " << element << endl;
	
	if( size == 0 ) {
		
		// DEBUGGING ..
		//cout << "replaceElement (size == 0) - replacePosition: " << replacePosition << endl;
		
		ptr = new T[1];									// create space for new Array, point to it ..
	
		assert(ptr != 0);								// terminate if memory not allocated ..
	
		size = 1;										// set size for this object ..
	
		ptr[0] = element;								// insert the value ..
		
	} else if( (0 <= replacePosition) && (replacePosition < size) ) {
		
		// DEBUGGING ..
		//cout << "replaceElement (replacePosition(" << replacePosition << ") < size(" << size << ")) - replacePosition: " << replacePosition << endl;

		ptr[replacePosition] = element;
		
		
	} // end "if( size == 0 )" ..
	
	
} // end "replaceElement( T element, int replacePosition )" ..

template<class T>
T& Array<T>:: removeElement( int removePosition ){
	
	// DEBUGGING ..
	//cout << "removeElement: " << removePosition << endl;
	
	if( (size > 0) && (0 <= removePosition) && (removePosition < size) ) {
		
		// DEBUGGING ..
		//cout << "removeElement (removePosition(" << removePosition << ") < size(" << size << ")) - removePosition: " << removePosition << endl;
		
		T* newPtr = new T[size-1];
		
		assert(newPtr != 0);
		
		// account for translating between previous array
		// to new array (following remove operation),
		// where new array has one less cell ..
		int removeSpace = 0;
		
		for(int i=0; i<size; i++){
			
			if ( i == removePosition ) {
				
				removeSpace++;
				
				// DEBUGGING ..
				//cout << "removeElement (removePosition < size): ptr[" << i << "](" << ptr[i] << ") removed." << endl;
				
			} else {
				
				// before insert operation: (i - insertSpace) = i - 0
				//  after insert operation: (i - insertSpace) = i - 1 ..
				
				newPtr[(i - removeSpace)] = ptr[i];
				
				// DEBUGGING ..
				//cout << "removeElement (removePosition < size): newPtr[" << (i-removeSpace) << "](" << newPtr[(i-removeSpace)] << ") = ptr[" << (i) << "](" << ptr[i] << ")" << endl;
				
			}
			
		} // end "for()" ..
		
		// update size of this Array ..
		--size;
		
		// update "ptr" with new values ..
		delete [] ptr;
		
		ptr = newPtr;
		
		
	} // end "if( size == 0 )" ..
	
} // end "removeElement( int removePosition )" ..


template<class T>
T& Array<T>:: getElement( int retrievePosition ){
	
	if( (size > 0) && (0 <= retrievePosition) && (retrievePosition < size) ) {
		
		return ptr[retrievePosition];
		
	}
	
} // end "getElement( int retrievePosition )" ..


template<class T>
void Array<T>:: pushElement( T element ){
	
	// DEBUGGING ..
	//cout << "pushElement - element: " << element << endl;
	
	if( size == 0 ) {
		
		// DEBUGGING ..
		//cout << "pushElement (size == 0): " << element << endl;
		
		ptr = new T[1];									// create space for new Array, point to it ..
	
		assert(ptr != 0);								// terminate if memory not allocated ..
	
		size = 1;										// set size for this object ..
	
		ptr[0] = element;								// insert the value ..
		
	} else {
		
		T* newPtr = new T[size+1];
		
		assert(newPtr != 0);
		
		++size;
		
		for(int i=0; i<size; i++){
			
			if ( i == (size-1) ) {						// "push" the last element ..
				
				newPtr[i] = element;
				
				// DEBUGGING ..
				//cout << "pushElement (last element): newPtr[" << i << "](" << newPtr[i] << ") = element(" << element << ")" << endl;
				
			} else {
				
				newPtr[i] = ptr[i];
				
				// DEBUGGING ..
				//cout << "pushElement (intermediate elem): newPtr[" << i << "](" << newPtr[i] << ") = ptr[" << i << "](" << ptr[i] << ")" << endl;
				
			}
			
		} // end "for()" ..
		
		
		// update "ptr" with new values ..
		delete [] ptr;
		
		ptr = newPtr;
		
		
	} // end "if( size == 0 )" ..
	
	
} // end "pushElement( T element )" ..


template<class T>
T& Array<T>:: popElement(){
	
	// DEBUGGING ..
	cout << "popElement .. "<< endl;
	
	if( size > 0 ) {
		
		T* newPtr = new T[size-1];
		
		assert(newPtr != 0);
		
		--size;												// ignore the last element ..
		
		for(int i=0; i<size; i++){
			
			newPtr[i] = ptr[i];
				
			// DEBUGGING ..
			cout << "popElement (intermediate elem): newPtr[" << i << "](" << newPtr[i] << ") = ptr[" << i << "](" << ptr[i] << ")" << endl;
				
			
		} // end "for()" ..
		
		// save the last element from the old Array, in order to return it ..
		T lastElement = ptr[size];
		
		// DEBUGGING ..
		cout << "popElement: "<< lastElement << endl;
		
		// update "ptr" with new values ..
		delete [] ptr;
		
		ptr = newPtr;
		
		// return the last element that was saved above ..
		return lastElement;
		
		
	} // end "if( size > 0 )" ..
	
	
} // end "popElement()" ..


/*
 * OVERLOADED OPERATORS ..
 */
 

// NOTE: for some reason, compiler doesn't recognize the overloaded definitions
// made in this class.  So "cout << testArray;" wont work ..
// should be ok now .. ?

template<class T> 
ostream& operator<<( ostream& output, const Array<T>& array ) {
//ostream& Array<T>:: operator<<( ostream& output, const Array<T>& array ) {
	
	if ( array.size == 0 ) {
		
		output << "Error outputting Array: Array is empty (size = " << array.size << ")" << endl;
		
	} else {
		
		for(int i=0; i<array.size; i++){
		
			output << "element[" << i << "]: " << array.ptr[i] << endl;
			
		}
		
	}
	
	return output;									// enables "output << x << y << z"
	
} // end "operator<<( ostream& output, const Array<T>& array )" ..


// NOTE: for some reason, compiler doesn't recognize the overloaded definitions
// made in this class.  So "cin >> testArray;" wont work ..
// should be ok now .. ?

template<class T> 
istream& operator>>( istream& input, Array<T>& array ) {
//istream& Array<T>:: operator>>( istream& input, Array<T>& array ) {
	
	for(int i=0; i<array.size; i++){
		
		cout << "element[" << i << "]: ";
		
		input >> array.ptr[i];
			
	}
	
	return input;									// enables "input >> x >> y >> z"
	
} // end "operator>>( istream& input, Array<T>& array )" ..

 
template<class T>
const Array<T>& Array<T>:: operator= ( const Array<T>& rValue ) {
			  // lvalue 	   = 			  rvalue
			
	if( &rValue != this ) {							// avoid "self-assignment" ..
		
		delete [] ptr;								// deallocate memory ..
		
		ptr = new T[rValue.size];					// create space for new Array, point to it ..
	
		assert(ptr != 0);							// terminate if memory not allocated ..
	
		size = rValue.size;							// set size for this object ..
	
		for(int i=0; i<size; i++){					// initialize new Array ..
		
			ptr[i] = rValue.ptr[i];					// set the values ..
			
		}
		
	}
	
	return *this;									// enables "x = y = z" assignment ..
	
} // end "operator= ( const Array<T>& rValue )" ..


template<class T>
bool Array<T>:: operator== ( const Array& rValue ) const {
	
	// check whether (size of rValue) == (size of Array[size]) ..
	if( size != rValue.size ){
	
		return false;								// sizes not equal, so Arrays NOT equal ..
		
	}
	
	// check whether rValue[i] == Array[i] ..
	for(int i=0; i<size; i++){
		
		if ( ptr[i] != rValue.ptr[i] )
			return false;							// elements not equal, so Arrays NOT equal ..
		
	}
	
	return true;									// both size and elements equal, so Arrays are equal ..
	
} // end "operator== ( const Array& rValue )" ..


template<class T>
bool Array<T>:: operator!= ( const Array& rValue ) const {
	
	// check whether (size of rValue) == (size of Array[size]) ..
	if( size != rValue.size ){
	
		return true;								// sizes not equal, so Arrays NOT equal ..
		
	}
	
	// check whether rValue[i] == Array[i] ..
	for(int i=0; i<size; i++){
		
		if ( ptr[i] != rValue.ptr[i] )
			return true;							// elements not equal, so Arrays NOT equal ..
		
	}
	
	return false;									// both size and elements equal, so Arrays are equal ..
	
} // end "operator!= ( const Array& rValue )" ..


template<class T>
T& Array<T>:: operator[] ( int subscript ) {
	
	assert( ((0 <= subscript) && (subscript < size)) );	// check for "subscript out of range" error ..
	
	return ptr[subscript];						// return reference for lvalue ..
	
} // end "operator[] (int subscript)" ..


/*
 * UTILITY FUNCTIONS ..
 */

template<class T>
void Array<T>:: print() const {
	
	if ( size == 0 ) {
		
		cout << "Error printing Array: Array is empty (size = " << size << ")" << endl;
		
	} else {
		
		for(int i=0; i<size; i++){
		
			cout << "element[" << i << "]: " << ptr[i] << endl;
			
		}
		
	}
	
} // end "print()" ..


template<class T>
void Array<T>:: print(char* prompt) const {
	
	if ( size == 0 ) {
		
		cout << "Error printing Array: Array is empty (size = " << size << ")" << endl;
		
	} else {
		
		for(int i=0; i<size; i++){
		
			cout << prompt << "[" << i << "]: " << ptr[i] << endl;
			
		}
		
	}
	
} // end "print()" ..


template<class T>
int Array<T>:: split( int first, int last ){
	
		// initialize some parameters ..
		T splitVal = ptr[first];
		
		bool onCorrectSide;
		
		int saveFirst = first;
		
		first++;
		
		do{
			
			onCorrectSide = true;
			
			// move "firstElement" toward "lastElement" ..
			while( onCorrectSide ){
				
				if( ptr[first] > splitVal )
					onCorrectSide = false;
				else {
					
					first++;
					onCorrectSide = ( first <= last );
					
				}
				
			}
			
			onCorrectSide = ( first <= last );
			
			// move "lastElement" toward "firstElement" ..
			while( onCorrectSide ){
				
				if( ptr[last] <= splitVal )
					onCorrectSide = false;
				else {
					
					last--;
					onCorrectSide = ( first <= last );
					
				}
				
			}
			
			if( first < last ){
				
				// swap "firstElement" and "lastElement" ..
				T temp = ptr[first];
				ptr[first] = ptr[last];
				ptr[last] = temp;
				
				first++;
				last--;
				
			}
			
		} while( first <= last );
		
		// swap "saveFirst" and "lastElement" ..
		T temp = ptr[saveFirst];
		ptr[saveFirst] = ptr[last];
		ptr[last] = temp;
		
		return last;
	
} // end "split( int firstElement, int lastElement )" ..


template<class T>
void Array<T>:: quicksort( int firstElement, int lastElement ){

	if ( firstElement < lastElement ) {
		
		// (1) calculate the "split point" ..
		
		int splitPoint = split( firstElement, lastElement );
		
		// (2) recursive calls to this function ..
		quicksort( firstElement, (splitPoint - 1) );
		quicksort( (splitPoint + 1), lastElement );
		
	} // end "if ( firstElement < lastElement )" ..


} // end "quicksort( int firstElement, int lastElement )" ..


template<class T>
void Array<T>:: sort(){
	
	int firstElement = 0;
	
	int lastElement = (size - 1);
	
	quicksort( firstElement, lastElement );
	
} // end "sort()" ..


template<class T>
int Array<T>:: reverse_split( int first, int last ){
	
		// initialize some parameters ..
		T splitVal = ptr[first];
		
		bool onCorrectSide;
		
		int saveFirst = first;
		
		first++;
		
		do{
			
			onCorrectSide = true;
			
			// move "firstElement" toward "lastElement" ..
			while( onCorrectSide ){
				
				if( ptr[first] <= splitVal )
					onCorrectSide = false;
				else {
					
					first++;
					onCorrectSide = ( first <= last );
					
				}
				
			}
			
			onCorrectSide = ( first <= last );
			
			// move "lastElement" toward "firstElement" ..
			while( onCorrectSide ){
				
				if( ptr[last] > splitVal )
					onCorrectSide = false;
				else {
					
					last--;
					onCorrectSide = ( first <= last );
					
				}
				
			}
			
			if( first < last ){
				
				// swap "firstElement" and "lastElement" ..
				T temp = ptr[first];
				ptr[first] = ptr[last];
				ptr[last] = temp;
				
				first++;
				last--;
				
			}
			
		} while( first < last );
		
		// swap "saveFirst" and "lastElement" ..
		T temp = ptr[saveFirst];
		ptr[saveFirst] = ptr[last];
		ptr[last] = temp;
		
		return last;
	
} // end "reverse_split( int firstElement, int lastElement )" ..


template<class T>
void Array<T>:: reverse_quicksort( int firstElement, int lastElement ){

	if ( firstElement < lastElement ) {
		
		// (1) calculate the "split point" ..
		
		int splitPoint = reverse_split( firstElement, lastElement );
		
		// (2) recursive calls to this function ..
		reverse_quicksort( firstElement, (splitPoint - 1) );
		reverse_quicksort( (splitPoint + 1), lastElement );
		
	} // end "if ( firstElement > lastElement )" ..


} // end "reverse_quicksort( int firstElement, int lastElement )" ..


template<class T>
void Array<T>:: reverse_sort(){
	
	int firstElement = 0;
	
	int lastElement = (size - 1);
	
	reverse_quicksort( firstElement, lastElement );
	
} // end "reverse_sort()" ..


template<class T>
void Array<T>:: merge(const Array<T>& secondArray) {
	
	int mergeSize = size + secondArray.size;			// calculate size of "merged" array ..
	
	T* mergePtr = new T[mergeSize];						// create space for new "mergeArray", point to it ..
	
	assert(mergePtr != 0);								// terminate if memory not allocated ..
	
	for(int i=0; i<size; i++){							// initialize new "mergeArray" with "ptr" portion ..
		
		mergePtr[i] = ptr[i];							// set the values ..
			
	}
	
	for(int j=0; j<secondArray.size; j++){				// initialize new "mergeArray" with "secondArray" portion ..
		
		mergePtr[(size+j)] = secondArray.ptr[j];		// set the values ..
			
	}
	
	delete [] ptr;										// deallocate space for "mergeArray" ..
	
	ptr = mergePtr;										// pointer points at new "mergeArray" ..
	
	size = mergeSize;									// set new size for this object ..
	
	sort();

} // end "merge(const Array<T>& secondArray)" ..


template<class T>
void Array<T>:: flip(){
	
	T* flipPtr = new T[size];							// create space for new "flipArray", point to it ..
	
	assert(flipPtr != 0);								// terminate if memory not allocated ..
	
	for(int i=0, j=(size-1); i<size, j>=0; i++, j--){	// initialize new "flipArray" ..
		
		flipPtr[i] = ptr[j];							// set the values ..
			
	}

	delete [] ptr;										// deallocate space for old Array ..
	
	ptr = flipPtr;										// pointer points at new "flipArray" ..
	
} // end "flip()" ..


template<class T>
int Array<T>:: search( T targetValue ) {
	
	int location = 0;
	
	bool found = false;
	
	bool moreToSearch = ( location < size );
	
	while( moreToSearch && !found ){
		
		if( ptr[location] == targetValue ){
			
			found = true;
			
		} else {
			
			location++;
			
			moreToSearch = ( location < size );
			
		}
		
	}
	
	if( found )
		return location;
	else
		return -1;										// return -1 if "targetValue" not found ..
	
} // end "search( T findThisValue )" ..


#endif /*ARRAY_H_*/